#!/usr/bin/env python
# coding: utf-8

import sys
import re
import logging
import traceback
import os
import warnings

import openpyxl
import openpyxl.utils
import time
import datetime
import json
import hashlib
import elasticsearch
import elasticsearch.helpers
from six import itervalues
# from werkzeug.utils import text_type
import psycopg2 as d_b_c
import psycopg2

INSERT_NORMAL = 0
INSERT_REPLACE = 1
INSERT_IGNORE = 2
import mysql.connector as d_b_c
import mysql.connector as mysql_connector
import pymysql
from mysql.connector.errors import DatabaseError

cur_e=(d_b_c.OperationalError, d_b_c.InterfaceError)
INSERT_MODES = {INSERT_IGNORE: 'insert ignore', INSERT_REPLACE: 'replace', INSERT_NORMAL: 'insert'}
ESCAPE="`%s`"
PY2 = sys.version_info[0] == 2
WIN = sys.platform.startswith('win')

# ESCAPE = '"%s"'
# cur_e = Exception
# INSERT_MODES = {INSERT_IGNORE: 'insert', INSERT_REPLACE: 'replace', INSERT_NORMAL: 'insert'}
"""
PostgreSQL 中简单实现 Insert ignore 操作
create 
    rule [规则名称] as on insert to [表名] 
where 
    exists 
        (select 1 from [表名] where [判断条件]) 
    do instead nothing; """

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.DEBUG)
handler = logging.StreamHandler(stream=sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    fmt='[%(levelname)s %(asctime)s.%(msecs)03d] [%(process)d:%(threadName)s:%(funcName)s:%(lineno)d] %(message)s',
    datefmt="%Y-%m-%d %H:%M:%S")
handler.setFormatter(formatter)
LOG.addHandler(handler)


def call_cost(func):
    def handler(*args, **kwargs):
        t = time.time()
        try:
            return func(*args, **kwargs)
        finally:
            t = round(time.time() - t, 3)
            logging.info(u'函数 {} 耗时 {} 秒'.format(func.__name__, t))

    return handler


class ExcelWriter:
    ILLEGAL_CHARACTERS_RE = re.compile(r'[\000-\010]|[\013-\014]|[\016-\037]')

    def __init__(self, filename=None):
        self.wb = None
        self.ws = None
        self.f_idx = 0
        self.filename = filename
        time_now = time.time()
        ms = int((time_now - int(time_now)) * 1000000)
        self.filename_auto = '{}-{}-{:06d}'.format(
            os.path.basename(sys.argv[0]).split(".")[0], datetime.datetime.now().strftime('%Y%m%d%H%M%S'), ms)
        self.new_file(self.filename)

    def write_row(self, row, indices=None):
        try:
            self.ws.append(row)
        except openpyxl.utils.exceptions.IllegalCharacterError as ex:
            if not indices:
                logging.warning('Failed to write excel: {}'.format(ex))
                return
            for index in indices:
                row[index] = re.sub(self.ILLEGAL_CHARACTERS_RE, '', row[index])
            try:
                self.ws.append(row)
            except Exception as ex:
                logging.warning('Failed to write excel: {}\n{}'.format(ex, traceback.format_exc()))

    @call_cost
    def new_file(self, filename=None):
        self.wb = openpyxl.Workbook()
        self.ws = self.wb.create_sheet(index=0)
        self.filename = filename or 'res-{}-{}.xlsx'.format(self.filename_auto, self.f_idx)

    @call_cost
    def save(self):
        self.wb.save(self.filename)
        logging.info('write excel fin: {}'.format(self.filename))

    @call_cost
    def write_all_excel(self, result):
        res = []
        for _ in result:
            res.append(_)
        result = res
        keys = [key for key in result[0].keys()]
        self.ws.append(keys)
        per_file_num = 0
        for idx, d in enumerate(result):
            if not idx % 50000:
                logging.info('write excel: {}/{} ..{:.2f}%'.format(idx, len(result), idx / (len(result) or 1) * 100))
            row = [str(d.get(key, '')) for key in keys]
            self.write_row(row)
            per_file_num += 1
            if per_file_num > 900000:
                self.save()
                self.f_idx += 1
                per_file_num = 0
                self.new_file()
                self.ws.append(keys)
        self.save()


class T:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


def secure_filename(filename):
    if isinstance(filename, str):
        from unicodedata import normalize
        filename = normalize('NFKD', filename).encode('utf-8', 'ignore')  # 转码
        if not PY2:
            filename = filename.decode('utf-8')  # 解码
    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, ' ')

    # 正则增加对汉字和日语假名的部分（本人有需求）
    # \ / : * ? " < > |    r"[\/\\\:\*\?\"\<\>\|]"
    # [ \\[ \\] \\^ \\-_*×――(^)$%~!@#$…&%￥—+=<>《》!！??？:：•`·、。，；,.;\"‘’“”-]
    # \u4E00-\u9FBF 中文
    # \u3040-\u30FF 假名
    # \u31F0-\u31FF 片假名扩展
    # _filename_ascii_add_strip_re = re.compile(r'[^A-Za-z0-9_\u4E00-\u9FBF\u3040-\u30FF\u31F0-\u31FF.-]')
    _filename_ascii_add_strip_re = re.compile(r"[\/\\\:\*\?\"\<\>\|]")

    filename = str(_filename_ascii_add_strip_re.sub('', '_'.join(  # 新的正则
        filename.split()))).strip('._')

    _windows_device_files = ('CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4', 'LPT1',
                             'LPT2', 'LPT3', 'PRN', 'NUL')

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if os.name == 'nt' and filename and \
       filename.split('.')[0].upper() in _windows_device_files:
        filename = '_' + filename

    return filename[:250]


def mkdir(file_name):
    path = os.path.join(get_realpath(), file_name)
    if not os.path.exists(path):
        os.mkdir(path)
    return path


def get_realpath():
    return os.path.split(os.path.realpath(__file__))[0]


class EsModel(object):
    def __init__(self, hosts, username=None, password=None, timeout=120):
        self.hosts = hosts
        self.username = username
        self.password = password
        if all([username, password]):
            http_auth = (self.username, self.password)
        else:
            http_auth = None
        self._es_client = elasticsearch.Elasticsearch(hosts=self.hosts, http_auth=http_auth, timeout=timeout)

    def get_info(self):
        return self.es.info()

    def analyze(self, query, analyzer='standard'):
        iclient = elasticsearch.client.indices.IndicesClient(self.es)
        return iclient.analyze(body={'text': query, 'analyzer': analyzer})

    def scan(
            self,
            query=None,
            scroll="30m",
            raise_on_error=True,
            preserve_order=False,
            size=1000,
            request_timeout=None,
            clear_scroll=True,
            scroll_kwargs=None,
            **kwargs
    ):
        """
        Simple abstraction on top of the
        :meth:`~elasticsearch.Elasticsearch.scroll` api - a simple iterator that
        yields all hits as returned by underlining scroll requests.

        By default scan does not return results in any pre-determined order. To
        have a standard order in the returned documents (either by score or
        explicit sort definition) when scrolling, use ``preserve_order=True``. This
        may be an expensive operation and will negate the performance benefits of
        using ``scan``.

        :arg client: instance of :class:`~elasticsearch.Elasticsearch` to use
        :arg query: body for the :meth:`~elasticsearch.Elasticsearch.search` api
        :arg scroll: Specify how long a consistent view of the index should be
            maintained for scrolled search
        :arg raise_on_error: raises an exception (``ScanError``) if an error is
            encountered (some shards fail to execute). By default we raise.
        :arg preserve_order: don't set the ``search_type`` to ``scan`` - this will
            cause the scroll to paginate with preserving the order. Note that this
            can be an extremely expensive operation and can easily lead to
            unpredictable results, use with caution.
        :arg size: size (per shard) of the batch send at each iteration.
        :arg request_timeout: explicit timeout for each call to ``scan``
        :arg clear_scroll: explicitly calls delete on the scroll id via the clear
            scroll API at the end of the method on completion or error, defaults
            to true.
        :arg scroll_kwargs: additional kwargs to be passed to
            :meth:`~elasticsearch.Elasticsearch.scroll`

        Any additional keyword arguments will be passed to the initial
        :meth:`~elasticsearch.Elasticsearch.search` call::

            scan(es,
                query={"query": {"match": {"title": "python"}}},
                index="orders-*",
                doc_type="books"
            )

        """
        scroll_kwargs = scroll_kwargs or {}

        if not preserve_order:
            query = query.copy() if query else {}
            query["sort"] = "_doc"

        # initial search
        resp = self.es.search(
            body=query, scroll=scroll, size=size, request_timeout=request_timeout, **kwargs
        )
        scroll_id = resp.get("_scroll_id")

        try:
            while scroll_id and resp["hits"]["hits"]:
                for hit in resp["hits"]["hits"]:
                    yield hit

                # check if we have any errors

                # if (resp["_shards"]["successful"] + resp["_shards"]["skipped"]) < resp["_shards"]["total"]:
                #     logging.warning(
                #         "Scroll request has only succeeded on %d (+%d skipped) shards out of %d.",
                #         resp["_shards"]["successful"],
                #         resp["_shards"]["skipped"],
                #         resp["_shards"]["total"],
                #     )
                #     if raise_on_error:
                #         raise Exception(
                #             scroll_id,
                #             "Scroll request has only succeeded on %d (+%d skiped) shards out of %d."
                #             % (resp["_shards"]["successful"], resp["_shards"]["skipped"], resp["_shards"]["total"]),
                #         )

                resp = self.es.scroll(
                    body={"scroll_id": scroll_id, "scroll": scroll}, **scroll_kwargs
                )
                scroll_id = resp.get("_scroll_id")

        finally:
            if scroll_id and clear_scroll:
                self.es.clear_scroll(body={"scroll_id": [scroll_id]}, ignore=(404,))

    # def scan(self, query, index, doc_type='_doc', scroll_id=None, scroll_total=0, scroll='30m',
    #          preserve_order=False, window=1000, size=None, offset=0, fields=None,
    #          clear_scroll=True, sort=None, sleep=0, query_task_id=None,
    #          **scroll_kwargs):
    #     query_str = str(json.dumps(query).replace("'", '"'))
    #     print(query_str)
    #     if not scroll_id:
    #         if not preserve_order:
    #             query = query.copy() if query else {}
    #             query["sort"] = "_doc"
    #
    #         # initial search
    #         resp = self.es.search(body=query, index=index, scroll=scroll, size=size, sort=sort)
    #
    #         scroll_id = resp.get('_scroll_id')
    #         scroll_total = resp.get('hits', {}).get('total')['value']
    #         first_run = True
    #
    #     else:
    #         first_run = False
    #         scroll_total = scroll_total
    #         resp = {}
    #     fin = 0
    #     if scroll_id is None:
    #         return
    #     try:
    #         idx  = 0
    #         while True:
    #             idx+=1
    #             # if we didn't set search_type to scan initial search contains data
    #             if first_run:
    #                 first_run = False
    #             else:
    #                 print(idx,scroll_id)
    #
    #                 resp = self.es.scroll(scroll_id=scroll_id)
    #                                       # params={'source': [] if fin < offset else fields},
    #                                       # source=[] if fin < offset else fields,)
    #             for hit in resp['hits']['hits']:
    #                 yield hit
    #             if size:
    #                 scroll_id = resp.get('_scroll_id')
    #                 print(resp, scroll_id)
    #             # check if we have any errrors
    #             # if resp["_shards"]["successful"] < resp["_shards"]["total"]:
    #             #     raise Exception(
    #             #         'Scroll request has only succeeded on %d shards out of %d.' % (
    #             #         resp['_shards']['successful'], resp['_shards']['total'])
    #             #     )
    #
    #             # end of scroll
    #             if scroll_id is None or not resp['hits']['hits'] or (
    #                     size and fin - offset >= size) or scroll_total == fin:
    #                 break
    #     finally:
    #         if scroll_id and clear_scroll:
    #             self.es.clear_scroll(body={'scroll_id': [scroll_id]}, ignore=(404,))

    def select(self, query, index, doc_type="_doc", fields=None, offset=0, limit=0, sort=None, result_format=True):
        offset = offset or 0
        limit = limit or 0
        if not limit:
            for record in elasticsearch.helpers.scan(self.es, index=index, doc_type=doc_type, query=query,
                                                     _source_include=fields, from_=offset, sort=sort):
                yield record['_source'] if result_format else record
        else:
            for record in self.es.search(index=index, doc_type=doc_type,
                                         body=query, _source_include=fields, from_=offset, size=limit,
                                         sort=sort
                                         ).get('hits', {}).get('hits', []):
                yield record['_source'] if result_format else record

    def search(self, query, index, doc_type="_doc", fields=None, offset=0, limit=0, sort=None):
        return self.es.search(index=index, doc_type=doc_type, body=query,
                              _source_include=fields, from_=offset, size=limit,
                              sort=sort)

    def msearch(self, query, index, doc_type="_doc"):
        body = '{}\n%s' % '\n{}\n'.join([json.dumps(q) for q in query])  # header\n body\n
        return self.es.msearch(index=index, doc_type=doc_type, body=body)['responses']

    def msearch_format(self, query, index, doc_type="_doc"):
        body = '{}\n%s' % '\n{}\n'.join([json.dumps(q) for q in query])  # header\n body\n
        return [r['hits']['hits'][0] for r in self.es.msearch(
            index=index, doc_type=doc_type, body=body)['responses'] if r['hits']['hits']]

    def count(self, index, doc_type="_doc", query=None):
        temp_query = {"query": query.get('query', {})}
        if not temp_query['query']:
            temp_query.pop('query')
        return self.es.count(index=index, doc_type=doc_type,
                             body=temp_query).get('count', 0)

    def get(self, es_id, index, doc_type="_doc", fields=None):
        ret = self.es.get(index=index, doc_type=doc_type, id=es_id,
                          _source_include=fields, ignore=404)
        return ret.get('_source', None)

    def drop(self, query, index, doc_type="_doc"):
        self.refresh(index)
        for record in elasticsearch.helpers.scan(self.es, index=index, doc_type=doc_type, query=query, _source=False):
            self.es.delete(index=index, doc_type=doc_type, id=record['_id'])

    def refresh(self, index):
        """
        Explicitly refresh one or more index, making all operations
        performed since the last refresh available for search.
        """
        self.es.indices.refresh(index=index)

    def copy(self):
        """
        Explicitly refresh one or more index, making all operations
        performed since the last refresh available for search.
        """
        return self

    @property
    def es(self):
        return self._es_client

    def bulk_write(self, data, batch_size=1000, retry=3):
        total = len(data)
        if retry < 0:
            retry = -retry
        elif retry > 10:
            retry = 10
        succeed = 0

        for _ in range(retry):
            try:
                result = elasticsearch.helpers.bulk(self.es, data, request_timeout=150, chunk_size=batch_size,
                                                    raise_on_error=False, raise_on_exception=False)
                succeed = result[0]
                if len(result) >1 and result[1]:
                    logging.error(result)
                if succeed == total:
                    return succeed
            except Exception as e:
                logging.error(e, traceback.format_exc())
        return succeed

    @staticmethod
    def _add_time_range(query, field='created_at', starttime=None, endtime=None):
        if starttime or endtime:
            time_range = {'range': {field: {'format': 'epoch_second'}}}
            if starttime:
                time_range['range'][field]['gte'] = starttime
            if endtime:
                time_range['range'][field]['lte'] = endtime
            query['query']['bool']['filter'].append(time_range)
        return query

    @staticmethod
    def _add_highlight(query):
        query['highlight'] = {
            "number_of_fragments": 0,  # fragment 是指一段连续的文字。返回结果最多可以包含几段不连续的文字。默认是5。
            "fragment_size": 0,  # 一段 fragment 包含多少个字符。默认100。
            "require_field_match": False,
            "pre_tags": "<font color=\"red\">",
            "post_tags": "</font>",
            "encoder": "html",
            "fields": {
                "*": {}
            }
        }
        return query

    @staticmethod
    def _add_aggregations_screen_name(query, offset=0, limit=2147483647):
        """
        "aggregations" : {
            "<aggregation_name>" : { <!--聚合的名字 -->
                "<aggregation_type>" : { <!--聚合的类型 -->
                    <aggregation_body> <!--聚合体：对哪些字段进行聚合 -->
                }
                [,"meta" : {  [<meta_data_body>] } ]? <!--元 -->
                [,"aggregations" : { [<sub_aggregation>]+ } ]? <!--在聚合里面在定义子聚合 -->
            }
            [,"<aggregation_name_2>" : { ... } ]*<!--聚合的名字 -->
        }

        聚合后分页详情见 https://blog.csdn.net/laoyang360/article/details/79112946

        :param query:
        :return:
        """
        query['size'] = 0
        query['from'] = offset
        query['aggs'] = {
            "group_agg": {
                "terms": {
                    "field": "user_screen_name",
                    "size": limit,
                    # 'from': offset,
                    "order": [{"_count": "desc"}],
                    # 广度搜索方式数据量越大，那么默认的使用深度优先的聚合模式生成的总分组数就会非常多，
                    # 但是预估二级的聚合字段分组后的数据量相比总的分组数会小很多所以这种情况下使用广度优先的模式能大大节省内存
                    "collect_mode": "breadth_first",
                },

                "aggs": {
                    "%s_%s" % ('user_screen_name', 'top_hits'): {
                        "top_hits": {
                            "size": 1,
                            "sort": [
                                {
                                    "created_at": {
                                        "order": "desc"
                                    }
                                }
                            ],
                            "_source": {
                                "includes": ['user_screen_name', 'text', 'id', 'retweeted']}}}}}}
        return query

    @staticmethod
    def _replace_highlight(data, keyword='text'):
        if keyword not in data['highlight']:
            return data
        if isinstance(data['highlight'][keyword], list):
            data['_source'][keyword] = ''.join(data['highlight'][keyword])
        else:
            data['_source'][keyword] = data['highlight'][keyword]
        return data


class ClientPyMySQL:
    INSERT_NORMAL = 'insert'
    INSERT_REPLACE = 'replace'
    INSERT_IGNORE = 'insert ignore'
    INSERT_MODES = {INSERT_IGNORE, INSERT_REPLACE, INSERT_NORMAL}
    PLACEHOLDER = '%s'

    def __init__(self, host, port, database,
                 user, passwd):
        # Connect to mysql.
        # dbc = pymysql.connect(
        #     host=Config.MYSQL_HOST, user=Config.MYSQL_USER,
        #     password=Config.MYSQL_PASSWORD, database=Config.MYSQL_DATABASE,
        #     port=Config.MYSQL_PORT,
        #     charset='utf8mb4',
        #     use_unicode=True,
        #     cursorclass=pymysql.cursors.DictCursor
        # )
        """
        SteadyDB
        DBUtils.SteadyDB基于兼容DB-API 2接口的数据库模块创建的普通连接，
        实现了"加强"连接。具体指当数据库连接关闭、丢失或使用频率超出限制时，将自动重新获取连接。

        典型的应用场景如下：在某个维持了某些数据库连接的程序运行时重启了数据库，
        或在某个防火墙隔离的网络中访问远程数据库时重启了防火墙。

        mincached:链接池中空闲链接的初始数量
        maxcached:链接池中空闲链接的最大数量
        maxshared:共享链接的最大数量
        maxconnections:建立链接池的最大数量
        blocking:超过最大链接数量时候的表现，为True等待链接数量降低，为false直接报错处理
        maxusage:单个链接的最大重复使用次数
        """

        import pymysql
       
        from dbutils.pooled_db import PooledDB
        self.dbc = PooledDB(
            creator=pymysql,
            host=host, user=user,
            passwd=passwd,
            database=database if database else None,
            port=port,
            charset='utf8mb4',
            use_unicode=True,
            autocommit=False,
            mincached=1, maxcached=10,
            maxshared=10, maxconnections=10, blocking=False,
            maxusage=None, setsession=None, reset=True,
            failures=None, ping=1,
            # 流式、字典访问
            cursorclass=pymysql.cursors.SSDictCursor)
        self.__pool = self.dbc
        self._conn = None
        self._cursor = None
        self.__get_conn()

    def __get_conn(self):
        self._conn = self.__pool.connection()
        self._cursor = self._conn.cursor()

    @property
    def dbcur(self):
        try:
            # if self.dbc.unread_result:
            #     self.dbc.get_rows()
            if not self._cursor:
                self.__get_conn()
            return self._cursor
        except (mysql_connector.OperationalError, mysql_connector.InterfaceError):
            self.__get_conn()
            self._cursor.ping(reconnect=True)
            return self._cursor

    # def begin_transaction(self):
    #     self.dbc.begin()
    #
    # def end_transaction(self):
    #     self.dbc.commit()

    def begin_transaction(self):
        return self.dbcur.execute("BE"+"GIN;")

    def end_transaction(self):
        return self.dbcur.execute("COMMIT;")

    def rollback(self):
        return self.dbcur.execute("rollback;")

    def fetch(self, dbcur):
        result = self.dbcur.fetchone()
        while result is not None:
            result = self.dbcur.fetchone()

    # def _execute1(self, sql, values=None):
    #     print(sql)
    #     with self.dbcur as dbcur:
    #         try:
    #             if values:
    #                 dbcur.execute(sql, values or [])
    #             else:
    #                 dbcur.execute(sql+';')
    #
    #             def _fetch(_dbcur):
    #                 result = _dbcur.fetchmany(2)
    #                 print(result)
    #                 while result is not None:
    #                     for r in result:
    #                         yield r
    #                     result = _dbcur.fetchmany(2)
    #             return dbcur.lastrowid, _fetch(dbcur)
    #         except mysql_connector.DatabaseError as ex:
    #             if ex.errno == 1205:
    #                 logging.critical(traceback.format_exc())
    #             else:
    #                 raise

    def _execute(self, sql, values=None, commit=False, **kwargs):
        print(sql)
        try:
            if values:
                self.dbcur.execute(sql, values or [])
            else:
                self.dbcur.execute(sql)
            if commit:
                self.dbcur.execute('COMMIT;')

            def _fetch():
                result =  self.dbcur.fetchmany(1000)
                while result:
                    # print(result)
                    for r in result:
                        yield r
                    result =  self.dbcur.fetchmany(1000)
            return self.dbcur.lastrowid, _fetch()
        except mysql_connector.DatabaseError as ex:
            if ex.errno == 1205:
                logging.critical(traceback.format_exc())
            else:
                raise

    @staticmethod
    def escape(string):
        return '`%s`' % string

    def save_into_db(self, sql, datas, batch_size=1000):
        for i in range((len(datas) + batch_size - 1) // batch_size):
            self._executemany(sql, datas[i * batch_size:(i + 1) * batch_size])

    def _executemany(self, sql_query, params):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # logging.debug("<sql: %s>" % sql_query)
            self.dbcur.execute("BE" + "GIN;")
            self.dbcur.executemany(sql_query, params)
            self.dbcur.execute("COMMIT;")
            return self.dbcur.fetchall()

    def _get_insert_sql(self, _mode=INSERT_NORMAL, _tablename=None, **values):
        if _mode not in self.INSERT_MODES:
            return None

        tablename = self.escape(_tablename)
        res_values = []
        res_keys = []
        if values:
            for k, v in values.items():
                res_keys.append(k)
                res_values.append(v)
            _keys = ", ".join((self.escape(k) for k in res_keys))
            _values = ", ".join([self.PLACEHOLDER, ] * len(values))
            sql_query = "%s INTO %s (%s) VALUES (%s)" % (_mode, tablename, _keys, _values)
        else:
            sql_query = "%s INTO %s DEFAULT VALUES" % (_mode, tablename)
        logging.debug("<sql: %s>", sql_query)
        return sql_query, res_keys, res_values

    def _insert(self, _mode=INSERT_NORMAL, _tablename=None, **values):
        sql, _, _ = self._get_insert_sql(_mode, _tablename, **values)

        if values:
            self.dbcur = self._execute(sql, list(itervalues(values)))
        else:
            self.dbcur = self._execute(sql)
        return self.dbcur.lastrowid

    def insert_many_with_dict_list(
            self, tablename, data,
            mode='INSERT IGNORE',
            # mode='REPLACE',
            batch_size=5000,
            *args,
            **kwargs
    ):
        if not data:
            return

        tablename = self.escape(tablename)
        values = data[0]
        if values:
            _keys = ", ".join((self.escape(k) for k in values))
            _values = ", ".join([self.PLACEHOLDER, ] * len(values))
            # sql_query = "%s INTO %s (%s) VALUES (%s)" % (mode, tablename, _keys, _values)
            sql_query = "%s INTO %s (%s) VALUES (%s)" % (mode, tablename, _keys, _values)
        else:
            # sql_query = "%s INTO %s DEFAULT VALUES" % (mode, tablename)
            sql_query = "%s INTO %s DEFAULT VALUES" % (mode, tablename)
        # self.logger.debug("<sql: %s>", sql_query)
        # self.logger.debug("%s", tuple(data[0][k] for k in values))

        size = max([1, (len(data) + batch_size - 1) // batch_size])
        for i in range(size):
            # self.logger.info("{} {}/{}".format(mode, batch_size * (i + 1), len(data)))
            self._executemany(sql_query, [list(d[k] for k in values) for d in data[i * batch_size:(i + 1) * batch_size]])
            # self.begin_transaction()
            # self._executemany(sql_query, [list(d[k] for k in values) for d in data[i * batch_size:(i + 1) * batch_size]])
            # self.end_transaction()
        return len(data)


class BaseDB(object):
    placeholder = '%s'

    def __init__(self, host, port, database,
                 user, passwd, logger=None, db_init=False):
        self.database = database
        self.host = host
        self.port = port
        self.user = user
        self.passwd = passwd
        self.logger = logger or LOG
        self.db_init = db_init
        self.conn = None
        self.init_db()

    def init_db(self):
        self.conn = d_b_c.connect(user=self.user, password=self.passwd,
                                  host=self.host, port=self.port,
                                  database=self.database,
                                  charset='utf8',
                                  # use_unicode=True,
                                  # autocommit=False,
                                  # 流式、字典访问
                                  # , autocommit=True
                                  )
        database = self.database
        try:
            print([x[0] for x in self._execute('show databases')])
            if database not in [x[0] for x in self._execute('show databases')]:
                self.conn.cursor().execute('CREATE DATABASE if not exists %s default charset utf8' % database)
            self.conn.database = self.database
        except cur_e:
            pass

        if self.db_init:
            self.create_db()

    def set_autocommit(self, autocommit):
        try:
            self.conn.set_autocommit(autocommit)
        except:
            self.conn.autocommit = autocommit

    def create_db(self):
        raise NotImplementedError

    def close(self):
        self.conn.commit()
        self.conn.close()

    @staticmethod
    def escape(string):
        # return '`%s`' % string
        return ESCAPE % string

    @property
    def dbcur(self):
        try:
            if self.conn.unread_result:
                self.conn.get_rows()
            return self.conn.cursor()
        except cur_e:
            try:
                self.conn.rollback()
            except:
                pass
            try:
                self.conn.ping(reconnect=True)
                self.conn.database = self.database
            except:
                pass
            return self.conn.cursor()

    def begin_transaction(self):
        self.dbcur.execute("BEGIN;")

    def end_transaction(self):
        self.dbcur.execute("COMMIT;")

    def rollback(self):
        self.dbcur.execute("rollback;")

    def _execute(self, sql_query, values=None, multi=False):
        if values is None:
            values = []
        dbcur = self.dbcur
        self.logger.debug("<sql: %s>" % sql_query)
        try:
            dbcur.execute(sql_query, values
                          # , multi=multi
                          )
            return dbcur
        except DatabaseError as ex:
            errno = ex.errno
            if errno == 1205:
                # 超过了锁定等待超时
                logging.critical(traceback.format_exc())
            elif errno in (1062, 1146):
                # 键重复条目(1062) 表不存在(1146)
                logging.warning(ex)
            else:
                raise

    def _executemany(self, sql_query, params, raise_on_error=True):
        dbcur = self.dbcur
        # self.logger.debug("<sql: %s>" % sql_query)
        try:
            dbcur.executemany(sql_query, params)
            return dbcur
        except DatabaseError as ex:
            if not raise_on_error and (ex.errno == 1205 or ex.errno == 1062):
                logging.critical(traceback.format_exc())
            else:
                raise

    def _select(self, tablename=None, what="*", where="", where_values=None,
                group_by=None, order_by=None, offset=0, limit=None):
        if where_values is None:
            where_values = []
        tablename = self.escape(tablename)
        if isinstance(what, list) or isinstance(what, tuple) or what is None:
            what = ','.join(self.escape(f) for f in what) if what else '*'

        sql_query = "SELECT %s FROM %s" % (what, tablename)
        if where:
            sql_query += " WHERE %s" % where
        if group_by:
            sql_query += " GROUP BY %s" % group_by
        if order_by:
            sql_query += " ORDER BY %s" % order_by
        if limit:
            sql_query += " LIMIT %d, %d" % (offset, limit)
        elif offset:
            sql_query += " LIMIT %d, %d" % (offset, limit)
        self.logger.debug("<sql: %s>" % sql_query)

        for row in self._execute(sql_query, where_values):
            yield row

    def _select2(self, tablename=None, what="*", where_columns=None, where_values=None,
                 group_by=None, order_by=None, offset=0, limit=None):
        if where_values is None:
            where_values = []
        if where_columns is None:
            where_columns = []
        tablename = self.escape(tablename)
        if isinstance(what, list) or isinstance(what, tuple) or what is None:
            what = ','.join(self.escape(f) for f in what) if what else '*'

        sql_query = "SELECT %s FROM %s" % (what, tablename)
        if where_columns:
            _condition_values = " and ".join([
                "%s = %s" % (self.escape(k), self.placeholder) for k in where_columns
            ])
            sql_query += " WHERE %s" % _condition_values
        if group_by:
            sql_query += " GROUP BY %s" % group_by
        if order_by:
            sql_query += " ORDER BY %s" % order_by
        if limit:
            sql_query += " LIMIT %d, %d" % (offset, limit)
        elif offset:
            sql_query += " LIMIT %d, %d" % (offset, limit)
        self.logger.debug("<sql: %s>" % sql_query)

        for row in self._execute(sql_query, where_values):
            yield row

    def insert(self, _mode=INSERT_NORMAL, _tablename=None, **values):
        if _mode not in (INSERT_NORMAL, INSERT_REPLACE, INSERT_IGNORE):
            return None
        _mode = INSERT_MODES[_mode]

        _tablename = self.escape(_tablename)
        if values:
            _keys = ", ".join((self.escape(k) for k in values))
            _values = ", ".join([self.placeholder, ] * len(values))
            sql_query = "%s INTO %s (%s) VALUES (%s)" % (_mode, _tablename, _keys, _values)
        else:
            sql_query = "%s INTO %s DEFAULT VALUES" % (_mode, _tablename)
        self.logger.debug("<sql: %s>", sql_query)

        if values:
            dbcur = self._execute(sql_query, list(itervalues(values)))
        else:
            dbcur = self._execute(sql_query)
        return dbcur.lastrowid if dbcur else 0

    def insert_many_with_dict_list(self, tablename, data, mode=INSERT_IGNORE, batch_size=1000):
        if not data or mode not in (INSERT_NORMAL, INSERT_REPLACE, INSERT_IGNORE):
            return

        mode = INSERT_MODES[mode]
        tablename = self.escape(tablename)
        values = data[0]
        if values:
            _keys = ", ".join((self.escape(k) for k in values))
            _values = ", ".join([self.placeholder, ] * len(values))
            # sql_query = "%s INTO %s (%s) VALUES (%s)" % (mode, tablename, _keys, _values)
            sql_query = "%s INTO %s (%s) VALUES (%s)" % (mode, tablename, _keys, _values)
        else:
            # sql_query = "%s INTO %s DEFAULT VALUES" % (mode, tablename)
            sql_query = "%s INTO %s DEFAULT VALUES" % (mode, tablename)
        # self.logger.debug("<sql: %s>", sql_query)
        # self.logger.debug("%s", tuple(data[0][k] for k in values))

        size = (len(data) + batch_size - 1) // batch_size
        for i in range(size):
            # self.logger.info("{} {}/{}".format(mode, batch_size * (i + 1), len(data)))
            # self._executemany(sql_query, [list(d[k] for k in values) for d in data[i * batch_size:(i + 1) * batch_size]])

            self._executemany(sql_query, [list(d[k] for k in values) for d in data[i * batch_size:(i + 1) * batch_size]])

        return len(data)

    def _update(self, _tablename=None, _where="1=0", _where_values=None, **values):
        if _where_values is None:
            _where_values = []
        _tablename = self.escape(_tablename)
        _key_values = ", ".join([
            "%s = %s" % (self.escape(k), self.placeholder) for k in values
        ])
        sql_query = "UPDATE %s SET %s WHERE %s" % (_tablename, _key_values, _where)
        self.logger.debug("<sql: %s>", sql_query)

        return self._execute(sql_query, list(itervalues(values)) + list(_where_values))

    def _update2(self, _tablename=None, _where_columns=None, _where_values=None, **values):
        if _where_values is None:
            _where_values = []
        if _where_columns is None:
            _where_columns = []
        _tablename = self.escape(_tablename)
        _key_values = ", ".join([
            "%s = %s" % (self.escape(k), self.placeholder) for k in values
        ])
        _condition_values = " and ".join([
            "%s = %s" % (self.escape(k), self.placeholder) for k in _where_columns
        ])
        if _where_columns:
            sql_query = "UPDATE %s SET %s WHERE %s" % (_tablename, _key_values, _condition_values)
            self.logger.debug("<sql: %s>", sql_query)
            return self._execute(sql_query, list(itervalues(values)) + list(_where_values))
        else:
            sql_query = "UPDATE %s SET %s" % (_tablename, _key_values)
            self.logger.debug("<sql: %s>", sql_query)
            return self._execute(sql_query, list(itervalues(values)))

    def _delete(self, tablename=None, where="1=0", where_values=None):
        if where_values is None:
            where_values = []
        tablename = self.escape(tablename)
        sql_query = "DELETE FROM %s" % tablename
        if where:
            sql_query += " WHERE %s" % where
        self.logger.debug("<sql: %s>", sql_query)

        return self._execute(sql_query, where_values)

    def __commit__(self):
        self.conn.commit()

    def __rollback__(self):
        self.conn.rollback()

    def __execute__(self, sql, vals=None):
        try:
            dbcur = self.dbcur
            dbcur.execute(sql, vals)
        except psycopg2.InternalError as e:
            if e.pgerror == 'ERROR:  current transaction is aborted, commands ignored until end of transaction block\n':
                self.__rollback__()
                return self.__execute__(sql, vals)
            raise e
        else:
            return dbcur

    def select(self, tab, query_dict, cols_selected='*', filter_='='):
        cols, vals = zip(*query_dict.items())
        cols_selected = ', '.join(cols_selected)
        expr_where = ' AND '.join(["{} {} %s".format(x, filter_) for x in cols])

        sql = "SELECT {} FROM \"{}\" WHERE {}".format(cols_selected, tab, expr_where)
        return self.__execute__(sql, vals)

        # --- omit ---

    def get(self, tab, query_dict, cols_selected='*', filter_='='):
        # here i need cursor
        self.cursor = self.select(tab, query_dict, cols_selected, filter_)
        if self.cursor.rowcount == 1:
            return self.cursor.fetchone()
        elif self.cursor.rowcount == 0:
            raise self.DoesNotExist("No record found!")
        else:
            raise self.MultipleObjectsReturned("Get {} records! only one excepted.".format(self.cursor.rowcount))


IP_STATUS_UNKNOWN = 0
IP_STATUS_ERRNETWORK = 1  # 网络故障
IP_STATUS_SUCCEED = 2  # 成功
IP_STATUS_EXCEED = 3  # IP重置超限


class NetworkException(Exception):
    pass


class ExceedException(Exception):
    pass





import os
from io import BufferedIOBase
import zipfile
from shutil import rmtree, move

from d22d.utils import get_file_md5


class _DisableSeekAndTellIOWrapper(BufferedIOBase):
    """
    zipfile压缩文件的时候会通过seek和tell来定位写入zip文件头信息。
    但是管道流不能正常的seek和tell，故通过这个wrapper来禁用seek和tell，从而强迫zipfile使用流式压缩
    """

    def __init__(self, towrap):
        self._wrapped = towrap

    def seek(self, *args, **kwargs):
        raise AttributeError()

    def tell(self, *args, **kwargs):
        raise AttributeError()

    def seekable(self, *args, **kwargs):
        return False

    def __getattribute__(self, item):
        if item in ('_wrapped', 'seek', 'seekable', 'tell'):
            return object.__getattribute__(self, item)
        return self._wrapped.__getattribute__(item)

    def __enter__(self):
        self._wrapped.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        self._wrapped.__exit__(*args, **kwargs)


def _zip_directory_to_pipe(dir_path: str) -> BufferedIOBase:
    rfd, wfd = os.pipe()

    def zip_in_thread(wfd):
        with os.fdopen(wfd, 'wb') as outputfile:
            with zipfile.ZipFile(_DisableSeekAndTellIOWrapper(outputfile), 'w') as zfile:
                for root, dirs, files in os.walk(dir_path):
                    for name in files:
                        fullpath = os.path.join(root, name)
                        print(fullpath)
                        arcname = os.path.relpath(fullpath, dir_path)
                        zfile.write(fullpath, arcname=arcname)

    import threading
    threading.Thread(target=zip_in_thread, args=(wfd,), name='zip ' + dir_path).start()
    return os.fdopen(rfd, 'rb')


def iter_zip_data(save_path, file_path):
    # import requests
    #
    # requests.post('http://localhost:8080/upload', files={'file': _zip_directory_to_pipe('d:/testdir')})
    with open(f'{save_path}.zip', 'wb') as wwf:
        cnt = 0
        while data := _zip_directory_to_pipe(file_path).read(8192):
            wwf.write(data)
            cnt += len(data)
            if not cnt % 10000:
                print(cnt)
                wwf.flush()


def zip_by_volume(file_path, block_size=1024 * 1024 * 50, ser='linux'):  # 分卷大小 50MB
    """zip单文件分卷压缩
    会删除文件所在目录名为：“文件名+_zip_split”的文件夹，再重新创建，在里面创建分卷
    分卷命名规则：末尾会用：“_分卷前文件MD5_分卷数.zip.001” -> “_分卷前文件MD5_分卷数.zip.999”
    返回目录路径和文件数量
    """
    file_size = os.path.getsize(file_path)  # 文件字节数
    count = file_size // block_size + 1

    folder_path, file_name = os.path.split(file_path)  # 除去文件名以外的path，文件名
    file_name_base = file_name.split('.')[0]  # 文件后缀名
    save_dir = os.path.join(folder_path, file_name) + '_zip_split'
    save_path = os.path.join(save_dir, file_name_base + f'_{int(count)}.zip')
    # 创建分卷压缩文件的保存路径
    if os.path.exists(save_dir):
        rmtree(save_dir)
    os.makedirs(save_dir)

    # 添加到临时压缩文件
    with zipfile.ZipFile(save_path, 'w') as zf:
        zf.write(file_path, arcname=file_name)
    md5_value = get_file_md5(save_path)
    # 小于分卷尺寸则直接返回压缩文件路径
    if file_size <= block_size:
        last_path = os.path.join(save_dir, file_name_base + f'_{md5_value}_1.zip')
        move(save_path, last_path)
        return save_dir, last_path, 1
    else:
        fp = open(save_path, 'rb')
        save_dir_last = f"{file_name_base}._{md5_value}_{count}.zip"
        if ser == 'win':
            for i in range(1, count + 1):
                _suffix = f'_{md5_value}_{count}.z{i:0>2}' if i != count else f'_{md5_value}_{count}.zip'
                name = os.path.join(save_dir, f"{file_name_base}.{_suffix}")
                f = open(name, 'wb+')
                if i == 1:
                    f.write(b'\x50\x4b\x07\x08')  # 添加分卷压缩header(4字节)
                    f.write(fp.read(block_size - 4))
                else:
                    f.write(fp.read(block_size))
        elif ser == 'linux':
            # 拆分压缩包为分卷文件
            for i in range(1, count + 1):
                _suffix = f'_{md5_value}_{count}.zip.{i:0>3}'
                name = os.path.join(save_dir, f"{file_name_base}.{_suffix}")
                f = open(name, 'wb+')
                if i == 1:
                    f.write(b'\x50\x4b\x07\x08')  # 添加分卷压缩header(4字节)
                    f.write(fp.read(block_size - 4))
                else:
                    f.write(fp.read(block_size))
        else:
            # 拆分压缩包为分卷文件
            for i in range(1, count + 1):
                _suffix = f'_{md5_value}_{count}.{i:0>3}' if i != count else f'_{md5_value}_{count}.zip'
                name = os.path.join(save_dir, f"{file_name_base}.{_suffix}")
                f = open(name, 'wb+')
                if i == 1:
                    f.write(b'\x50\x4b\x07\x08')  # 添加分卷压缩header(4字节)
                    f.write(fp.read(block_size - 4))
                else:
                    f.write(fp.read(block_size))
        fp.close()
        os.remove(save_path)  # 删除临时的 zip 文件
        return save_dir, save_dir_last, count


if __name__ == '__main__':
    file = r"/home/user/P.mp4"  # 原始文件
    path = zip_by_volume(file, ser='linux')
    print(path)  # 输出分卷压缩文件的路径、文件名、文件数



# /usr/bin/env python
# coding: utf-8


all_line = set([])
all_line_3 = set([])
all_line__ = set([])
all_line_base = set([])


def req_conflit(path):
    with open(path, 'r', encoding='utf8') as rf, \
            open(path + '.txt', 'w', encoding='utf8') as wf, \
            open(path.rsplit('/', maxsplit=1)[0] + '/setup.new.txt', 'w', encoding='utf8') as wsf:
        for line in rf:
            line = line.strip()
            if not line.strip():
                continue
            if line.startswith('#'):
                all_line_3.add(line)
            else:
                if line.find('=') > -1:
                    if line.find('>') > -1:
                        all_line__.add(line)
                        all_line.add(line.split('>')[0])
                    elif line.find('<') > -1:
                        all_line__.add(line)
                        all_line.add(line.split('<')[0])
                    else:
                        all_line__.add(line)
                        all_line.add(line.split('=')[0])
                elif line not in all_line:
                    all_line.add(line)
                    all_line_base.add(line)
        for line in all_line_base:
            wf.write('''{}\n'''.format(line.replace('\n', '')))
            wsf.write('''"{}",\n'''.format(line.replace('\n', '')))
        for line in all_line__:
            wf.write('''{}\n'''.format(line.replace('\n', '')))
            wsf.write('''"{}",\n'''.format(line.replace('\n', '')))
        for line in all_line_3:
            wf.write('''{}\n'''.format(line.replace('\n', '')))


def main():
    setting = r'../requirements.txt'
    req_conflit(setting)


if __name__ == '__main__':
    main()



import hashlib
import ntpath
import random
import string
import sys
import time
import os
import logging

import six
import wrapt
import traceback
import datetime
import shutil
import asyncio.exceptions

from functools import lru_cache
from tornado.log import LogFormatter as _LogFormatter

from sys import platform
from shutil import copy2
from json import JSONEncoder, dumps
from asyncio import sleep
from base64 import b64encode
from async_timeout import timeout as a_timeout

logger = logging.getLogger('d22d.task')


def dump_json(obj):
    if (not obj) or isinstance(obj, type(None)):
        return ''

    return dumps(obj, cls=JSONEncoderWithBytes)


def run_task_auto_retry(
        func, args=(), kwargs=None,
        warning_d=None,
        error_d=None,
        raise_e=None,
        time_sleep=5,
        max_retry=0,
        timeout=0,
):
    """

    :param timeout: 任务超时时间
    :param max_retry: 任务最大重试次数
    :param time_sleep: 错误等待时长
    :param func: 任务函数
    :param args: args参数
    :param kwargs: kwargs参数
    :param warning_d: 报警异常类 以及 异常信息 kv字典
    {Exception: "这里是报警信息"}
    :param error_d: 错误异常类 以及 异常信息 kv字典
    {Exception: "这里是错误信息"}
    :param raise_e: 需要抛出异常的异常类
    [Exception]
    :return:
    """
    if not isinstance(args, tuple):
        args = (args,)
    if not kwargs:
        kwargs = {}
    if not warning_d:
        warning_d = {}
    if not error_d:
        error_d = {}
    if not raise_e:
        raise_e = []
    retry = 0
    time_start = time.time()
    func_info = f'{func.__code__.co_filename}:{func.__code__.co_firstlineno}:{func.__name__}()'
    while True:
        retry += 1
        try:
            logger.debug(
                f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info}' +
                # f' {args} {kwargs}'+
                ''
            )
            res = func(*args, **kwargs)
            logger.info(
                f'[SUC:{retry:04d}:{time.time() - time_start:.2f}s] {func_info}' +
                # f' {args} {kwargs}'+
                ''
            )
            return res
        except tuple(warning_d.keys()) as ex:
            error = ex
            ex_msg = warning_d[ex.__class__]
            if ex_msg is not None:
                logger.warning(f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info} {ex_msg}')
        except tuple(error_d.keys()) as ex:
            error = ex
            ex_msg = error_d[ex.__class__]
            if ex_msg is not None:
                logger.error(f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info} {ex_msg}')
            if ex.__class__ in raise_e:
                raise ex
        except tuple({Exception: ""}.keys()) as ex:
            error = ex
            logger.error(f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s]  {func_info}'
                         f' {type(ex)} {str(ex)[:250]:250s}\n{traceback.format_exc()}')
            if ex.__class__ in raise_e:
                raise ex
        if (
                (max_retry and retry > max_retry) or
                (timeout and ((time.time() - time_start) > timeout))
        ):
            raise error
        time.sleep(time_sleep)


def task_auto_retry(
        warning_d=None,
        error_d=None,
        raise_e=None,
        time_sleep=3,
        max_retry=10,
        timeout=0,
):
    if not warning_d:
        warning_d = {}
    if not error_d:
        error_d = {}
    if not raise_e:
        raise_e = []
    """
      :param timeout: 任务超时时间
      :param max_retry: 任务最大重试次数
      :param time_sleep: 错误等待时长
      :param func: 任务函数
      :param args: args参数
      :param kwargs: kwargs参数
      :param warning_d: 报警异常类 以及 异常信息 kv字典
      {Exception: "这里是报警信息"}
      :param error_d: 错误异常类 以及 异常信息 kv字典
      {Exception: "这里是错误信息"}
      :param raise_e: 需要抛出异常的异常类
      [Exception]
      :return:
    """

    @wrapt.decorator
    def wrapper(func, _instance, args, kwargs):
        if not isinstance(args, tuple):
            args = (args,)
        if not kwargs:
            kwargs = {}
        retry = 0
        time_start = time.time()
        func_info = f'{func.__code__.co_filename}:{func.__code__.co_firstlineno}:{func.__name__}()'
        while True:
            retry += 1
            try:
                logger.debug(
                    f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info}' + ''
                    # f' {args} {kwargs}'
                )
                return func(*args, **kwargs)
            except tuple(warning_d.keys()) as ex:
                error = ex
                ex_msg = warning_d[ex.__class__]
                if ex_msg is not None:
                    logger.warning(f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info} {ex_msg}')
            except tuple(error_d.keys()) as ex:
                error = ex
                ex_msg = error_d[ex.__class__]
                if ex_msg is not None:
                    logger.error(f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info} {ex_msg}')
                if ex.__class__ in raise_e:
                    raise ex
            except tuple({Exception: ""}.keys()) as ex:
                error = ex
                logger.error(
                    f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s]  {func_info}'
                    f' {type(ex)} {str(ex)[:250]:250s}\n{traceback.format_exc()}',
                    # f' {args} {kwargs}',
                )
                if ex.__class__ in raise_e:
                    raise ex
            if (
                    (max_retry and retry > max_retry) or
                    (timeout and ((time.time() - time_start) > timeout))
            ):
                raise error
            time.sleep(time_sleep)

    return wrapper


def async_task_auto_retry(
        warning_d=None,
        error_d=None,
        raise_e=None,
        time_sleep=5,
        max_retry=0,
        timeout=0,
        hard_timeout=0,
        # callback=None,
        # error_callback=None,
        finally_callback=None
):
    if not warning_d:
        warning_d = {}
    if not error_d:
        error_d = {}
    if not raise_e:
        raise_e = []

    """
      :param timeout: 任务超时时间
      :param hard_timeout: 任务超时时间退出
      :param max_retry: 任务最大重试次数
      :param time_sleep: 错误等待时长
      :param func: 任务函数
      :param args: args参数
      :param kwargs: kwargs参数
      :param warning_d: 报警异常类 以及 异常信息 kv字典
      {Exception: "这里是报警信息"}
      {Exception: lambda ex: f'这里是报警信息{ex.x}'}
      :param error_d: 错误异常类 以及 异常信息 kv字典
      {Exception: "这里是错误信息"}
      :param raise_e: 需要抛出异常的异常类
      [Exception]
      :return:
    """

    @wrapt.decorator
    async def wrapper(func, _instance, args, kwargs):
        nonlocal hard_timeout
        nonlocal timeout
        if callable(hard_timeout):
            _hard_timeout = hard_timeout(*args, **kwargs)
        else:
            _hard_timeout = hard_timeout
        if callable(timeout):
            _timeout = timeout(*args, **kwargs)
        else:
            _timeout = timeout

        if not isinstance(args, tuple):
            args = (args,)
        if not kwargs:
            kwargs = {}
        retry = 0
        time_start = time.time()
        func_info = f'{func.__code__.co_filename}:{func.__code__.co_firstlineno}:{func.__name__}()'
        while True:
            retry += 1
            try:
                logger.debug(
                    f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info}' + ''
                    # f' {args} {kwargs}'
                )

                try:
                    if _hard_timeout:
                        async with a_timeout(_hard_timeout - int(time.time() - time_start)) as _:
                            result = await func(*args, **kwargs)
                            # print(_.expired)
                    else:
                        result = await func(*args, **kwargs)
                except asyncio.exceptions.TimeoutError:
                    e_t = TimeoutError(f'{func_info}超时 {_hard_timeout}s')
                    setattr(e_t, 'seconds', _hard_timeout)
                    raise e_t
                return result
            except tuple(warning_d.keys()) as ex:
                error = ex
                ex_msg = warning_d[ex.__class__]
                if ex_msg is not None:
                    if callable(ex_msg):
                        ex_msg = ex_msg(ex)
                    logger.warning(f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info} {ex_msg}')
                if (raise_e is True) or (ex.__class__ in raise_e):
                    raise ex
            except tuple(error_d.keys()) as ex:
                error = ex
                ex_msg = error_d[ex.__class__]
                if ex_msg is not None:
                    if callable(ex_msg):
                        ex_msg = ex_msg(ex)
                    logger.error(f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info} {ex_msg}')
                    if (raise_e is True) or (ex.__class__ in raise_e):
                        raise ex
            except tuple({Exception: ""}.keys()) as ex:
                error = ex
                import traceback
                logger.error(f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s]  {func_info}'
                             f' {type(ex)} {str(ex)[:150]}'
                             f'\n{traceback.format_exc()}',
                             # f' {args} {kwargs}',
                             )
                if (raise_e is True) or (ex.__class__ in raise_e):
                    raise ex
            finally:
                if finally_callback is not None:
                    finally_callback(*args, **kwargs)
            if (
                    (max_retry and retry > max_retry) or
                    (_timeout and ((time.time() - time_start) > _timeout))
            ):
                raise error
            await sleep(time_sleep)

    return wrapper


def custom_time(timestamp):
    time_local = time.localtime(timestamp)
    dt = time.strftime("%Y-%m-%d", time_local)
    return dt


class JSONEncoderWithBytes(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bytes):
            return b64encode(obj).decode('ascii')
        elif hasattr(obj, 'to_json'):
            return obj.to_json()
        elif hasattr(obj, 'to_dict'):
            return dumps(obj.to_dict(), cls=JSONEncoderWithBytes)
        elif hasattr(obj, 'isoformat'):
            # datetime
            return obj.isoformat()
        elif isinstance(obj, datetime.datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')
        # elif hasattr(obj, '__str__'):
        #     return str(obj)
        elif isinstance(obj, datetime.date):
            return obj.strftime('%Y-%m-%d')
        else:
            try:
                return super().default(obj)
            except Exception as ex:
                if hasattr(obj, '__str__'):
                    return str(obj)
                elif hasattr(obj, '__repr__'):
                    return obj.__repr__()
                else:
                    raise ex


def format_error(ex):
    return '[{}] {}\n{}'.format(type(ex), ex, traceback.format_exc())


def utf8(string):
    """
    Make sure string is utf8 encoded bytes.

    If parameter is a object, object.__str__ will been called before encode as bytes
    """
    if isinstance(string, six.text_type):
        return string.encode('utf8')
    elif isinstance(string, six.binary_type):
        return string
    else:
        return six.text_type(string).encode('utf8')


def get_md5(s):
    md5 = hashlib.md5(utf8(s)).hexdigest()
    return str(md5)


def get_file_md5(file):
    md5 = hashlib.md5()
    f = open(file, 'rb')
    md5.update(f.read())
    f.close()
    return str(md5.hexdigest()).lower()


class LogFormatter(_LogFormatter, object):
    def __init__(self, fmt=None, datefmt=None, color=True, *args, **kwargs):
        datefmt = "%Y-%m-%d %H:%M:%S"
        # '[%(threadName)s]' \
        if fmt == 1:
            fmt = '%(color)s[%(asctime)s][%(levelname)1.1s]%(end_color)s%(message)s'
        self.fmt = fmt or '%(color)s[%(asctime)s][%(levelname)1.1s][%(name)s][%(filename)s:%(lineno)d:%(funcName)s()]%(end_color)s%(message)s'

        super(LogFormatter, self).__init__(color=color, fmt=self.fmt, datefmt=datefmt, *args, **kwargs)

    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            s = str(datetime.datetime.now().strftime(datefmt))
        else:
            t = time.strftime("%Y-%m-%d %H:%M:%S", ct)
            s = "%s.%03d" % (t, record.msecs)
        return s


logger_info_where = logging.getLogger('info_where')
logger_info_where.setLevel(logging.DEBUG)


def set_shell_log(log, fmt=None):
    # create a file handler
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)

    # create a logging format
    handler.setFormatter(LogFormatter(fmt=fmt))

    # add the handlers to the logger
    log.addHandler(handler)
    if log.name != 'root':
        log.propagate = False


def log_info(*args, **kwargs):
    # 什么函数调用了此函数
    which_fun_call_this = sys._getframe(1).f_code.co_name  # NOQA
    # 获取被调用函数在被调用时所处代码行数
    line = sys._getframe().f_back.f_lineno
    # 获取被调用函数所在模块文件名
    file_name = sys._getframe(1).f_code.co_filename
    logger_info_where.info(f'"{file_name}:{line}" {which_fun_call_this}()：' + ' '.join(
        [arg.__str__() for arg in args] + [', ', ', '.join(f"{k}={v}" for k, v in kwargs.items())]))


def active_log():
    # set_shell_log(logger)
    set_shell_log(logging.getLogger())
    set_shell_log(logger_info_where, 1)


def activate_debug_logger(level=logging.DEBUG):
    """Global logger used when running from command line."""
    logging.basicConfig(
        format='(%(levelname)s) %(message)s', level=level
    )


def with_cur_lock():
    @wrapt.decorator
    def wrapper(func, instance, args, kwargs):
        lock = instance.cur_lock
        if lock.locked():
            raise IOError(f'{args[0:]} cur locked')
        lock.acquire()
        try:
            res = func(instance.cur, *args, **kwargs)
            return res
        except Exception as ex:
            raise ex
        finally:
            lock.release()

    return wrapper


def gen_pass(val_type='all', val_len=8):
    src_digits = string.digits  # string_数字  '0123456789'
    src_uppercase = string.ascii_uppercase  # string_大写字母 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    src_lowercase = string.ascii_lowercase  # string_小写字母 'abcdefghijklmnopqrstuvwxyz'
    src_special = string.punctuation  # string_特殊字符 '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'

    if val_type.lower().find('int') == -1:
        # sample从序列中选择n个随机独立的元素，返回列表
        num = random.sample(src_digits, 1)  # 随机取1位数字
        lower = random.sample(src_uppercase, 1)  # 随机取1位小写字母
        upper = random.sample(src_lowercase, 1)  # 随机取1位大写字母
        special = random.sample(src_special, 1)  # 随机取1位大写字母特殊字符
        other = random.sample(string.ascii_letters + string.digits + string.punctuation, val_len - 4)
        # 生成字符串
        # print(num, lower, upper, special, other)
        pwd_list = num + lower + upper + special + other
        # shuffle将一个序列中的元素随机打乱，打乱字符串
        random.shuffle(pwd_list)
        # 列表转字符串
        password_str = ''.join(pwd_list)
        # print(password_str)
    else:
        pwd_list = random.sample(string.digits, val_len)
        random.shuffle(pwd_list)
        # 列表转字符串
        password_str = int(''.join(pwd_list))

    return password_str


def remove_folder(path):
    if os.path.exists(path):
        shutil.rmtree(path)


def makedirs(path):
    folder_name = path.split('\\')[-1].split('/')[-1]
    is_folder = '.' not in folder_name
    path = os.path.realpath(path)
    out_f_path = os.path.dirname(path)
    if is_folder:
        out_f_path = os.path.join(out_f_path, folder_name)
    if not os.path.exists(out_f_path):
        logger.info(f'正在创建新文件夹：{out_f_path}，因为{path}需要')
        os.makedirs(out_f_path)
        

def iter_path(path, exclude=None, include=None, exclude_path=None, include_path=None, return_type=2, open_kwargs=None):
    if exclude is None:
        exclude = []
    if include is None:
        include = []
    if exclude_path is None:
        exclude_path = []
    if include_path is None:
        include_path = []
    if open_kwargs is None:
        open_kwargs = {}
    logging.info(f'开始遍历文件夹：{path}')
    cnt = 0
    cnt_size = 0
    cnt_iter = 0
    for root, fs, fns in os.walk(path):
        for fn in fns:
            cnt += 1
            if include and not any([bool(include_str in fn) for include_str in include]):
                continue
            if exclude and any([bool(exclude_str in fn) for exclude_str in exclude]):
                continue
            f_path = os.path.join(root, fn)
            if include_path and not any([bool(include_str in f_path) for include_str in include_path]):
                continue
            if exclude_path and any([bool(exclude_str in f_path) for exclude_str in exclude_path]):
                continue
            cnt_size += os.path.getsize(f_path)
            cnt_iter += 1
            logging.debug(f'遍历到文件：{f_path} [已处理{cnt_iter}|返回{cnt_iter}个：{cnt_size / 1024 / 1024:.3f}MB]')
            if return_type == 0:
                yield cnt, root, *os.path.splitext(fn)
            elif return_type == 1:
                yield cnt, root, fn
            elif return_type == 2:
                yield cnt, f_path
            elif return_type == 3:
                yield f_path
            elif return_type == 4:
                yield cnt, open(f_path, 'rb', **open_kwargs)
            elif return_type == 5:
                yield open(f_path, 'rb', **open_kwargs)
    logging.info(f'遍历文件结束 [{cnt_size / 1024 / 1024:.3f}MB] [已处理{cnt_iter}|返回{cnt_iter}个]：{path}')


def path_without_leaf(path):
    result = ""
    try:
        head, tail = ntpath.split(path)
        result = head
    except Exception as inst:
        logger.exception(f"path_without_leaf Error - {path}: {inst}")
    return result


# return the name of the file + extension
def path_leaf(path):
    result = ""
    try:
        head, tail = ntpath.split(path)
        result = tail or ntpath.basename(head)
    except Exception as inst:
        logger.exception(f"Path_leaf Error - {path}: {inst}")
    return result


def create_temporary_copy(path, temp_dir, file_name):
    if platform == "win32":
        file_to_check = temp_dir + '\\' + file_name
    else:
        file_to_check = temp_dir + '/' + file_name
    # if the file already exists, we will delete it
    exists = os.path.isfile(file_to_check)
    if exists:
        os.chmod(file_to_check, 777)
        os.remove(file_to_check)
    temp_path = os.path.join(temp_dir, file_name)
    copy2(path, temp_path)
    return temp_path


def time_stamp():
    """returns a formatted current time/date"""
    import time
    return str(time.strftime("%Y_%m_%d_%H_%M_%S"))



import os

from d22d.model import midhardware


class FileSystemStore(midhardware.BaseStore):
    def __init__(self, location):
        self.location = location

    def count_data(self):
        return NotImplementedError

    def list_data(self):
        return NotImplementedError

    def get_data(self, position, *args, **kwargs):
        return NotImplementedError

    def save_data(self, position, *args, **kwargs):
        return NotImplementedError

    def delete_data(self, position, *args, **kwargs):
        return NotImplementedError

    def get_position(self, position, *args, **kwargs):
        return os.path.join(self.location, position)

    def get_data_size(self, position, *args, **kwargs):
        return NotImplementedError



#!/usr/bin/env python3
# coding: utf-8
import csv
import os
import random
import re
import time
import traceback
import openpyxl.utils.exceptions
import openpyxl
import xlrd
import json
import logging
import pymongo
import datetime
import sqlite3
import pandas as pd
import copy
import re
import gridfs
import abc
from bson import ObjectId
import pymongo.errors
# import bson.objectid
from abc import ABC
from pathlib import Path
from .myutils import EsModel
from .myutils import BaseDB
from .myutils import ClientPyMySQL
from .myutils import ExcelWriter
from .sqlfileextra import SqlExtractor, match_insert, RANDOM_STR
from ..utils import format_error, with_cur_lock, gen_pass, run_task_auto_retry
from clickhouse_driver import connect as clickhouse_connect
from threading import Lock
from collections import defaultdict
from threading import local as threading_local


def get_table_name_from_sql(sql):
    pattern = {
    'select' : r"(?!')*\bfrom\b\s+\b(?P<table_name>\w+)\b\s+\b(?P<table_sn>\w+)\b(?!')*",
    'insert' : r"(?!')*\binsert\b\s+\binto\b\s+\b(?P<table_name>\w+)\b\s+\b(?P<table_sn>\w+)\b(?!')*",
    'update' : r"(?!')*\bupdate\b\s+\b(?P<table_name>\w+)\b\s+\b(?P<table_sn>\w+)(?!')+",
    }
    res = re.compile(pattern[sql.lower().strip().split(' ')[0]], re.S | re.IGNORECASE).search(sql)
    if res:
        return res['table_name']


def get_realpath():
    return os.path.split(os.path.realpath(__file__))[0]


def secure_filename(filename):
    if isinstance(filename, str):
        from unicodedata import normalize
        filename = normalize('NFKD', filename).encode('utf-8', 'ignore').decode('utf-8')  # 转码

    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, ' ')

    # 正则增加对汉字和日语假名的部分（本人有需求）
    # \ / : * ? " < > |    r"[\/\\\:\*\?\"\<\>\|]"
    # [ \\[ \\] \\^ \\-_*×――(^)$%~!@#$…&%￥—+=<>《》!！??？:：•`·、。，；,.;\"‘’“”-]
    # \u4E00-\u9FBF 中文
    # \u3040-\u30FF 假名
    # \u31F0-\u31FF 片假名扩展
    # _filename_ascii_add_strip_re = re.compile(r'[^A-Za-z0-9_\u4E00-\u9FBF\u3040-\u30FF\u31F0-\u31FF.-]')
    _filename_ascii_add_strip_re = re.compile(r"[\/\\\:\*\?\"\<\>\|/]")

    filename = str(_filename_ascii_add_strip_re.sub('', '_'.join(  # 新的正则
        filename.split()))).strip('._')

    _windows_device_files = ('CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4', 'LPT1',
                             'LPT2', 'LPT3', 'PRN', 'NUL')

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if os.name == 'nt' and filename and \
            filename.split('.')[0].upper() in _windows_device_files:
        filename = '_' + filename

    return filename[:250]


class ElasticSearchD(EsModel):
    def __init__(self, hosts, username=None, password=None, cols_ddl=None, ik=False):
        super().__init__(hosts, username, password)
        self.hosts = hosts
        self.username = username
        self.password = password
        self.ik = ik
        self.cols_ddl = cols_ddl

    def __repr__(self):
        return f'ElasticSearch:{self.hosts}'

    def get_data(self, index, *args, **kwargs):
        if isinstance(index, str):
            if 'condition' in kwargs:
                query = kwargs.pop('condition')
            else:
                query = {}
        else:
            query = index[1]
            index = index[0]
        logging.debug(json.dumps(query))
        for i in self.scan(query=query, index=index, *args, **kwargs):
            r = {}
            if r.get('id'):
                r['_id'] = i['_id']
            else:
                r['id'] = i['_id']
            r.update(i['_source'])
            yield r

    def save_data(self, index, data, batch_size=1000, pks='', pop_id='_id', *args, **kwargs):
        actions = []
        write_kws = {k: kwargs[k] for k in ['batch_size', 'retry'] if k in kwargs}
        for d in data:
            if pks:
                _id = '-'.join(d[k] for k in pks.split(','))
            elif '_id' in d:
                _id = d['_id']
            else:
                _id = '_'.join(map(lambda x: str(x) if len(str(x)) <= 10 else str(x)[0:11], d.values()))

            if pop_id:
                for k in pop_id.split(','):
                    d.pop(k)
            actions.append({
                '_op_type': 'index',  # 操作 index update create delete
                '_index': index,  # index
                '_type': '_doc',  # type
                "_id": f"{_id}",
                '_source': d})
            if len(actions) > batch_size:
                self.bulk_write(actions, **write_kws)
                actions = []

        if len(actions):
            self.bulk_write(actions, **write_kws)


    def get_indexes(self):
        return list(self.es.indices.get_alias().keys())

    def get_ddl(self, index):
        return self.es.indices.get_mapping(index)

    def get_count(self, index, *args, **kwargs):
        if isinstance(index, str):
            query = {}
        else:
            query = {'query': index[1]['query']}
            index = index[0]
            if query.get("_source"):
                query.pop("_source")
        return int(self.es.count(index=index, body=query)['count'])

    @staticmethod
    def get_int_type_from_len(length):
        if 0 < length <= 8:
            return "byte"
        elif 8 < length <= 16:
            return "short"
        elif 16 < length <= 32:
            return "integer"
        elif 32 < length <= 64:
            return "long"
        return None

    @staticmethod
    def get_str_type_from_len(length):
        if 0 < length <= 256:
            return "keyword"
        elif 256 < length <= 65536:
            return "text"
        elif 65536 < length <= 16777216:
            return "text"
        elif 16777216 < length <= 4294967295:
            return "text"
        return None

    def create_index(self, index, data, pks='id'):
        # sql = """SET NAMES utf8mb4;"""
        # sql += """SET FOREIGN_KEY_CHECKS = 0;"""
        # if drop == 1:
        #     sql += """DROP TABLE IF EXISTS `{}`;""".format(tbname)
        pairs = {
            "settings": {
                "index": {
                    "number_of_shards": 1,
                    "number_of_replicas": 1
                }
            },
            'mappings': {"doc": {
                'properties': {

                }}
            }
        }
        properties = pairs['mappings']['doc']['properties']
        # ES7以上删掉doc层
        # pairs = {
        #     "settings": {
        #         "index": {
        #             "number_of_shards": 1,
        #             "number_of_replicas": 1
        #         }
        #     },
        #     'mappings': {
        #         'properties': {
        #
        #         }
        #     }
        # }
        # properties = pairs['mappings']['properties']

        if self.cols_ddl:
            if self.cols_ddl[index].get('mappings'):
                pairs['mappings'] = self.cols_ddl[index]['mappings']
            else:
                for col, col_type in self.cols_ddl[index].items():
                    properties[col] = {}
                    if col_type in ['int', 'bigint', 'integer', 'tinyint', 'smallint', 'mediumint']:
                        properties[col]['type'] = 'long'
                    elif col_type in ['float', 'double']:
                        properties[col]['type'] = 'float'
                    elif col_type in ['bit']:
                        properties[col]['type'] = 'long'
                    elif col_type in ['char', 'varchar', 'text', 'tinyblob', 'tinytext', 'blob', 'mediumtext', 'mediumblob',
                                      'longtext', 'longblob', 'json', 'timestamp', 'datetime']:
                        properties[col] = {
                            "type": "text",
                            "fields": {
                                "keyword": {
                                    "type": "keyword",
                                    "ignore_above": 1024
                                }
                            },
                        }
                        if self.ik:
                            properties[col]["fields"]["analyzer"] = "ik_max_word"
                            properties[col]["fields"]["search_analyzer"] = "ik_smart"
        else:
            for name, value in data.items():
                if isinstance(value, dict):
                    continue
                properties[name] = {}
                if isinstance(value, int):
                    properties[name]['type'] = 'long'
                elif isinstance(value, float):
                    properties[name]['type'] = 'float'
                elif isinstance(value, str) or isinstance(value, list):
                    properties[name] = {
                        "type": "text",
                        "fields": {
                            "keyword": {
                                "type": "keyword",
                                "ignore_above": 1024
                            }
                        },
                    }
                    if self.ik:
                        properties[name]["fields"]["analyzer"] = "ik_max_word"
                        properties[name]["fields"]["search_analyzer"] = "ik_smart"
        if self.es.indices.exists(index=index) is not True:
            res = self.es.indices.create(index=index, body=pairs)
            if not res:
                logging.error("错误，创建索引{}失败".format(index))
                return False

            logging.info("正常，创建索引{}成功".format(index))
            return True
        else:
            logging.info("正常，索引{}已存在".format(index))
            return True


# class MySqlD(BaseDB, ABC):
#     def __init__(self, host, port, user, passwd, database=None):
#         super().__init__(host, port, database, user, passwd)
#
#     def get_data(self, index, *args, **kwargs):
#         return self.__execute__(sql=f'select * from {index}', *args, **kwargs).fetchall()
#
#     def save_data(self, index, *args, **kwargs):
#         self.insert_many_with_dict_list(tablename=index, *args, **kwargs)
#
#     def get_indexes(self):
#         tbs = []
#         for r in list(self.__execute__(sql='show tables;')):
#             tbs.append(r[0])
#         return tbs
#
#     def get_count(self, index):
#         return int(self.__execute__(sql=f'select count(0) as c from {index}').fetchone()[0])
#
#     @staticmethod
#     def get_int_type_from_len(length):
#         if 0 < length <= 8:
#             return "tinyint"
#         elif 8 < length <= 16:
#             return "smallint"
#         elif 16 < length <= 32:
#             return "int"
#         elif 32 < length <= 64:
#             return "bigint"
#         return None
#
#     @staticmethod
#     def get_str_type_from_len(length):
#         if 0 < length <= 256:
#             return "varchar({})".format(length)
#         elif 256 < length <= 65536:
#             return "text"
#         elif 65536 < length <= 16777216:
#             return "mediumtext"
#         elif 16777216 < length <= 4294967295:
#             return "longtext"
#         return None
#
#     def create_index(self, index, data, pks='id'):
#         # sql = """SET NAMES utf8mb4;"""
#         # sql += """SET FOREIGN_KEY_CHECKS = 0;"""
#         # if drop == 1:
#         #     sql += """DROP TABLE IF EXISTS `{}`;""".format(tbname)
#         sql = """create table if not exists {}(""".format(index)
#         for name, value in data.items():
#             if isinstance(value, int):
#                 _type = 'bigint'
#             elif isinstance(value, str):
#                 if pks.find(name) > -1:
#                     _type = 'varchar(256)'
#                 else:
#                     _type = 'text'
#             else:
#                 _type = 'blob'
#
#             sql += ('`' + name + '` ')
#             sql += (_type + ',')
#
#         slen = len(pks)
#         if slen > 0:
#             sql += """PRIMARY KEY ("""
#             v_keys = pks.split(',')
#             first = 1
#             for key in v_keys:
#                 if first == 1:
#                     first = 0
#                     sql += """`{}`""".format(key)
#                 else:
#                     sql += """,`{}`""".format(key)
#             sql += """)) """
#         else:
#             sql = sql[0: -1]
#             sql += ')'
#
#         sql += """ENGINE = InnoDB CHARACTER SET = utf8mb4 COLLATE = utf8mb4_general_ci ROW_FORMAT = Dynamic;"""
#         print(sql)
#         self.__execute__(sql)
#         self.__commit__()


class MySqlD(ClientPyMySQL, ABC):
    def __init__(self, host, port, user, passwd, database=None, tables_ddl=None):
        super().__init__(host, port, database, user, passwd)
        self.host = host
        self.port = port
        self.database = database
        self.tables_ddl = tables_ddl

    def __repr__(self):
        return f'MySQL:{self.host}:{self.port}/{self.database}'

    def get_data(self, index, *args, **kwargs):
        sub_sql = f"{index} {kwargs['condition']}" if 'condition' in kwargs else index
        if index.lower().strip().startswith('select '):
            return self._execute(sql=f'{sub_sql}', *args, **kwargs)[1]
        else:
            return self._execute(sql=f'select * from {sub_sql}', *args, **kwargs)[1]

    def save_data(self, index, data, *args, **kwargs):
        self.insert_many_with_dict_list(tablename=index, data=data, *args, **kwargs)

    def get_indexes(self):
        return list(t[f'Tables_in_{self.database}'] for t in self._execute(sql='show tables;')[1])

    def get_columns_type(self, index):
        sql = f"select table_name,column_name,data_type,column_comment,column_default,column_key " \
              f"from information_schema.columns " \
              f"where table_schema='{self.database}' and table_name='{index}';"
        return list(t for t in self._execute(sql=sql)[1])

    def get_table_ddl(self, index):
        sql = f"show create table {index}"
        return 'CREATE TABLE IF NOT EXISTS ' + self._execute(sql=sql)[1].__next__()['Create Table'].replace('\n', '').strip('CREATE TABLE')

    def get_count(self, index, *args, **kwargs):
        if index.lower().strip().startswith('select '):
            index = f'({index.lower().strip()})'
        for r in self._execute(f'select count(0) as c from  {index} as taSDFEWVempTABlesdfecH', )[1]:
            return r['c']

    @staticmethod
    def get_int_type_from_len(length):
        if 0 < length <= 8:
            return "tinyint"
        elif 8 < length <= 16:
            return "smallint"
        elif 16 < length <= 32:
            return "int"
        elif 32 < length <= 64:
            return "bigint"
        return None

    @staticmethod
    def get_str_type_from_len(length):
        if 0 < length <= 256:
            return "varchar({})".format(length)
        elif 256 < length <= 65536:
            return "text"
        elif 65536 < length <= 16777216:
            return "mediumtext"
        elif 16777216 < length <= 4294967295:
            return "longtext"
        return None

    def create_index(self, index, data, pks='id'):
        if self.tables_ddl:
            sql = self.tables_ddl[index]
            self._execute(sql, commit=True)
        else:
            # sql = """SET NAMES utf8mb4;"""
            # sql += """SET FOREIGN_KEY_CHECKS = 0;"""
            # if drop == 1:
            #     sql += """DROP TABLE IF EXISTS `{}`;""".format(tbname)
            sql = """create table if not exists {}(""".format(index)
            for name, value in data.items():
                if isinstance(value, int):
                    _type = 'bigint'
                elif isinstance(value, str):
                    if name in pks.split(','):
                        _type = 'varchar(256)'
                    else:
                        _type = 'text'
                else:
                    _type = 'blob'

                sql += ('`' + name + '` ')
                sql += (_type + ',')

            slen = len(pks)
            if slen > 0:
                sql += """PRIMARY KEY ("""
                v_keys = pks.split(',')
                first = 1
                for key in v_keys:
                    if first == 1:
                        first = 0
                        sql += """`{}`""".format(key)
                    else:
                        sql += """,`{}`""".format(key)
                sql += """)) """
            else:
                sql = sql[0: -1]
                sql += ')'

            sql += """ENGINE = InnoDB CHARACTER SET = utf8mb4 COLLATE = utf8mb4_general_ci ROW_FORMAT = Dynamic;"""
            print(sql)
            self._execute(sql, commit=True)


def get_line_num_fast(filename):
    count_n = 0
    count_r = 0
    fp = open(filename, "rb")
    while 1:
        buffer = fp.read(1 * 1024 * 1024)
        if not buffer:
            break
        count_n += buffer.count(b'\n')
        count_r += buffer.count(b'\r')
    fp.close()
    return count_n or count_r


class BaseFileD(object):
    def __init__(self, path, extension, encoding='utf8-sig', newline=None):
        self.path = path
        self.extension = extension
        self.encoding = encoding
        self.newline = newline
        self._file_w = dict()
        self._indexes_path = dict()

    def __repr__(self):
        return f'{self.extension}:{self.path}'

    def __del__(self):
        for f in self._file_w.values():
            try:
                f.close()
            except Exception as e:
                logging.warning(e)

    def gen_path_by_index(self, index):
        if index not in self._indexes_path:
            path = f'{self.path}{os.sep}{secure_filename(index)}.{self.extension}'
            self._indexes_path[index] = path
        return self._indexes_path[index]

    def get_indexes(self):
        res = []
        real_path = os.path.split(os.path.realpath(self.path))[0]
        for root, fs, fns in os.walk(self.path):
            for fn in fns:
                path = f'{root}{os.pathsep}{fn}'
                (filepath, tempfilename) = os.path.split(path)
                (filename, extension) = os.path.splitext(tempfilename)
                if extension not in [f'.{self.extension}']:
                    continue
                index = '-'.join(list(filepath.replace(real_path, '', 1).split(os.sep)) + [filename])
                res.append(index)
                self._indexes_path[index] = path
        return res

    def get_count(self, index, *args, **kwargs):
        return get_line_num_fast(self.gen_path_by_index(index))

    @classmethod
    def w_open_func(cls, *args, **kwargs):
        return open(*args, **kwargs)

    def create_index(self, index, data, pks='id'):
        if not os.path.exists(self.path):
            try:
                os.makedirs(self.path)
            except Exception as e:
                logging.warning(e)
                Path(self.path).mkdir(parents=True, exist_ok=True)
        path = self.gen_path_by_index(index)
        if os.path.exists(path):
            os.rename(path, f"{path}.{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}.bak")
        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        self._file_w.setdefault(index, self.w_open_func(path, 'w', encoding=self.encoding, newline=self.newline))


class CsvD(BaseFileD):
    def __init__(self, path, split=',', extension='csv', encoding='utf8'):
        super().__init__(path, extension, encoding, newline='')
        self.split = split
        self._file_w = dict()
        self.___file_w = dict()

    def get_count(self, index, *args, **kwargs):
        count = get_line_num_fast(self.gen_path_by_index(index))
        if count:
            count -= 1
        return count

    @property
    def __file_w(self):
        for k, v in self._file_w.items():
            if k not in self.___file_w:
                self.___file_w[k] = csv.DictWriter(v, fieldnames=[])
        return self.___file_w

    # @classmethod
    # def w_open_func(cls, *args, **kwargs):
    #     return csv.writer(open(*args, **kwargs))

    def get_data(self, index, fieldnames=None, restkey=None, restval=None,
                 dialect="excel", **kwargs):
        with open(self.gen_path_by_index(index), 'r', encoding=self.encoding) as f:
            for line in csv.DictReader(
                    f, fieldnames=fieldnames, restkey=restkey, restval=restval, dialect=dialect, **kwargs):
                yield line
            # keys = list(k.replace("'", '').replace('"', '') for k in f.readline().strip().split(self.split))
            # for line in f:
            #     yield {keys[idx]: v for idx, v in enumerate(eval(line))}

    def save_data(self, index, data, *args, **kwargs):
        # self._file_w[index].writelines((self.split.join(f'{v.__repr__()}' for v in d.values()) + '\n') for d in data)
        self.__file_w[index].writerows([d for d in data])
        self._file_w[index].flush()

    def create_index(self, index, data, pks='id'):
        super(self.__class__, self).create_index(index, data)
        # self._file_w[index].writerow((self.split.join(f'"{v.__repr__()[1:-1]}"' for v in data.keys()) + '\n'))
        self.__file_w[index].fieldnames = [v for v in data.keys()]
        self.__file_w[index].writeheader()
        self._file_w[index].flush()


class TxtD(BaseFileD):
    def __init__(self, path, split=',', extension='txt', encoding='utf8'):
        super().__init__(path, extension, encoding, newline='')
        self.split = split
        self._file_w = dict()
        self.___file_w = dict()

    def get_count(self, index, *args, **kwargs):
        count = get_line_num_fast(self.gen_path_by_index(index))
        if count:
            count -= 1
        return count

    @property
    def __file_w(self):
        for k, v in self._file_w.items():
            if k not in self.___file_w:
                self.___file_w[k] = v
        return self.___file_w

    def get_data(self, index, fieldnames=None, restkey=None, restval=None,
                 dialect="excel", **kwargs):
        with open(self.gen_path_by_index(index), 'r', encoding=self.encoding) as f:
            keys = list(k.strip().replace("'", '').replace('"', '') for k in f.readline().strip().split(self.split))
            for line in f:
                values = list(v.strip().replace("'", '').replace('"', '') for v in line.strip().split(self.split))
                yield {keys[idx]: v for idx, v in enumerate(values)}

    def save_data(self, index, data, *args, **kwargs):
        self._file_w[index].writelines((self.split.join(v.__repr__() for v in d.values()) + '\n') for d in data)
        self._file_w[index].flush()

    def create_index(self, index, data, pks='id'):
        super(self.__class__, self).create_index(index, data)
        self._file_w[index].writelines((self.split.join(v.__repr__() for v in data.keys()) + '\n'))
        self._file_w[index].flush()


class ZipD(object):
    def __init__(self, path, get_file_data_func=None, fieldnames=None, extension='zip', encoding='utf8'):
        self.path = path
        self.fieldnames = fieldnames or []
        self.extension = extension
        if get_file_data_func:
            self.get_file_data_func = get_file_data_func

    def get_file_data_func(self, file_path):
        with open(file_path, 'r', encoding='utf-8') as f:
            for line in f:
                yield {k: v for k, v in zip(self.fieldnames, line.strip().split(','))}

    def get_count(self, index, *args, **kwargs):
        return 1

    def get_data(self, index, **kwargs):
        from ..ziputils import un_zip, iter_path, remove_folder

        uz_path = None
        try:
            uz_path = un_zip(os.path.join(self.path, f'{index}.{self.extension}'), **kwargs)
            for file_path in iter_path(uz_path, return_type=3):
                for line in self.get_file_data_func(file_path):
                    yield line
        finally:
            if uz_path:
                remove_folder(uz_path)

    def save_data(self, index, data, *args, **kwargs):
        raise NotImplementedError

    def create_index(self, index, data, pks='id'):
        raise NotImplementedError


class RarD(ZipD):
    def __init__(self, path, get_file_data_func=None, fieldnames=None, extension='rar', encoding='utf8'):
        super().__init__(
            path=path, get_file_data_func=get_file_data_func,
            fieldnames=fieldnames or [], extension=extension, encoding=encoding)

    def get_data(self, index, **kwargs):
        from ..rarutils import un_rar, remove_folder
        from ..ziputils import iter_path

        ur_path = None
        try:
            ur_path = un_rar(os.path.join(self.path, f'{index}.{self.extension}'), **kwargs)
            for file_path in iter_path(ur_path, return_type=3):
                for line in self.get_file_data_func(file_path):
                    yield line
        finally:
            if ur_path:
                remove_folder(ur_path)


class SqlFileD(BaseFileD):
    def __init__(self, path, extension='sql', encoding='utf8', mode='insert', compress=False):
        super().__init__(path, extension, encoding)
        self.compress = compress
        self.mode = {
            1: 'INSERT',
            "insert": 'INSERT',
            "INSERT": 'INSERT',
            2: 'INSERT IGNORE',
            'insert ignore': 'INSERT IGNORE',
            'INSERT IGNORE': 'INSERT IGNORE',
            3: 'REPLACE',
            "replace": 'REPLACE',
            "REPLACE": 'REPLACE',
        }.get(mode, mode)

    def get_count(self, index, *args, **kwargs):
        count = get_line_num_fast(self.gen_path_by_index(index))
        if count:
            count -= 1
        return count

    def get_data(self, index):
        with open(self.gen_path_by_index(index), 'r', encoding=self.encoding) as f:
            table_info_str = ''
            flag_create_table = False
            table_keys = []

            def is_insert(li):
                return li.lower()[:7] in ('insert ', 'replace')

            def get_table_keys(table_info_str_raw):
                if table_info_str_raw.endswith(';'):
                    table_info_str_raw = table_info_str_raw[-1]
                info_str = f'{table_info_str_raw});'
                return list(SqlExtractor(info_str).get_v_main()[0].keys())

            for line in f:
                line = line.strip()
                if line.lower().startswith('create table'):
                    flag_create_table = True
                elif line.lower() == '-- ----------------------------':
                    flag_create_table = False
                    if table_info_str and not table_keys:
                        table_keys = get_table_keys(table_info_str)
                elif is_insert(line):
                    flag_create_table = False
                    if table_info_str and not table_keys:
                        table_keys = get_table_keys(table_info_str)
                if flag_create_table:
                    if line.strip().startswith('`') or line.strip().lower().startswith('create table'):
                        table_info_str += line
                if is_insert(line):
                    # for data in SqlExtractor(line.strip()).get_v_main(table_keys):
                    #     yield data
                    tbname, ks, vss = match_insert(line)
                    for vs in vss:
                        yield {k: v for k, v in zip(ks or table_keys, vs)}

        # con = sqlite3.connect(f'{self.gen_path_by_index(index)}')
        # query = f'select * from {index}'
        # data = pd.read_sql(query, con)
        # for data in data.to_dict():
        #     print(data)
        #     yield data
        # con.close()

    def save_data(self, index, data, update=None, *args, **kwargs):
        if self.compress:
            d_keys = ', '.join('`{}`'.format(data[0].keys()))
            self._file_w[index].write(
                '{} INTO `{}`({}) VALUES ({}){};\n'.format(
                    self.mode,
                    index,
                    d_keys,
                    '), ('.join(
                        ', '.join(
                            v.__str__() if isinstance(v, int) else (
                                'NULL' if isinstance(v, type(None)) else (
                                    "'{}'".format(
                                        v.__str__().replace(
                                            '\\"', RANDOM_STR).replace(
                                            "\\", '\\\\').replace(
                                            "'", '\\\'').replace(
                                            RANDOM_STR, '\\"')) if isinstance(v, str) else (
                                        f"{v.__str__().__repr__()}"
                                    )
                                )
                            )
                            for v in d.values()
                        )
                        for d in data
                    ),
                    f' {update}' if update else ''
                )
            )
        else:
            self._file_w[index].writelines(
                '{} INTO `{}`({}) VALUES ({}){};\n'.format(
                    self.mode,
                    index,
                    ', '.join('`{}`'.format(dk) for dk in d.keys()),
                    ', '.join(
                        v.__str__() if isinstance(v, int) else (
                            'NULL' if isinstance(v, type(None)) else (
                                "'{}'".format(v.__str__().replace(
                                    '\\"', RANDOM_STR).replace(
                                    "\\", '\\\\').replace(
                                    "'", '\\\'').replace(
                                    RANDOM_STR, '\\"')
                                ) if isinstance(v, str) else (
                                    f"{v.__str__().__repr__()}"
                                )
                            ))
                        for v in d.values()
                    ),
                    f' {update}' if update else ''
                )
                for d in data)

        self._file_w[index].flush()

    def create_index(self, index, data, pks='id'):
        super(self.__class__, self).create_index(index, data)
        # self._file_w[index].writerow((self.split.join(f'"{v.__repr__()[1:-1]}"' for v in data.keys()) + '\n'))


class JsonListD(BaseFileD):
    def __init__(self, path, extension='json', encoding='utf8', ensure_ascii=False):
        super().__init__(path, extension, encoding)
        self.ensure_ascii = ensure_ascii

    def get_data(self, index):
        with open(self.gen_path_by_index(index), 'r', encoding=self.encoding) as f:
            for line in f:
                yield json.loads(line.strip())

    def save_data(self, index, data, *args, **kwargs):
        self._file_w[index].writelines((json.dumps(d, ensure_ascii=self.ensure_ascii) + '\n') for d in data)
        self._file_w[index].flush()


class XlsIbyFileD(BaseFileD):
    ILLEGAL_CHARACTERS_RE = re.compile(r'[\000-\010]|[\013-\014]|[\016-\037]')

    def __init__(self, path, extension='xls'):
        self.path = path
        self.extension = extension
        self._file_w = {}
        self._file_line_num = {}
        self._file_w_now_sheet = {}
        self._file_w_now_sheet_num = {}
        self._file_w_now_sheet_line_num = {}
        self._file_w_now_keys = {}
        self._indexes_path = {}
        super().__init__(path, extension)

    def get_count(self, index, *args, **kwargs):
        all_line = 0
        workbook = xlrd.open_workbook(self.gen_path_by_index(index))
        for idx, name in enumerate(workbook.sheet_names()):
            logging.info(f'sheet:{idx}:{name}')
            worksheet = workbook.sheet_by_index(idx)
            all_line += worksheet.nrows
        return all_line

    def get_data(self, index):
        workbook = xlrd.open_workbook(self.gen_path_by_index(index))  # 文件路径
        # 获取所有sheet的名字
        for idx, name in enumerate(workbook.sheet_names()):
            logging.info(f'sheet:{idx}:{name}')
            worksheet = workbook.sheet_by_index(idx)
            nrows = worksheet.nrows
            if not nrows:
                continue
            keys = {i: key for i, key in enumerate(worksheet.row_values(0))}
            for line_num in range(1, nrows):
                line = worksheet.row_values(line_num)
                # shengri = worksheet.row(line_num)[3].ctype
                # print(shengri)
                data = {keys[i]: key for i, key in enumerate(line)}
                yield data

    def save_data(self, index, data, *args, **kwargs):
        for d in data:
            if not d:
                continue
            row = [str(d.get(key, '')) for key in self._file_w_now_keys[index]]
            self.write_row(self._file_w_now_sheet[index], row)
            self._file_w_now_sheet_line_num[index] += 1
            if self._file_w_now_sheet_line_num[index] > 500000:
                self._file_w_now_sheet_num[index] += 1
                self._file_w_now_sheet[index] = self._file_w[index].create_sheet(
                    index=self._file_w_now_sheet_num[index])
                keys = [key for key in d.keys()]
                self.write_row(self._file_w_now_sheet[index], keys)
                self._file_w_now_sheet_line_num[index] = 1
                self._file_w_now_keys[index] = keys

    @classmethod
    def write_row(cls, ws, row, indices=None):
        try:
            ws.append(row)
        except openpyxl.utils.exceptions.IllegalCharacterError as ex:
            if not indices:
                logging.warning('Failed to write excel: {}'.format(ex))
                return
            for index in indices:
                row[index] = re.sub(cls.ILLEGAL_CHARACTERS_RE, '', row[index])
            try:
                ws.append(row)
            except Exception as ex:
                logging.warning('Failed to write excel: {}\n{}'.format(ex, traceback.format_exc()))

    @classmethod
    def w_open_func(cls, *args, **kwargs):
        wb = openpyxl.Workbook()
        return wb

    def create_index(self, index, data, pks='id'):
        super().create_index(index, data)
        self._file_w_now_sheet_num[index] = 0
        self._file_w_now_sheet[index] = self._file_w[index].create_sheet(index=self._file_w_now_sheet_num[index])
        keys = [key for key in data.keys()]
        self.write_row(self._file_w_now_sheet[index], keys)
        self._file_w_now_sheet_line_num[index] = 1
        self._file_w_now_keys[index] = keys

    def __del__(self):
        for idx, f in self._file_w.items():
            while True:
                try:
                    start_time = time.time()
                    f.save(self._indexes_path[idx])
                    logging.info(
                        '[use {:.2f}s] write excel fin: {}'.format(time.time() - start_time, self._indexes_path[idx]))
                    break
                except ImportError as e:
                    break
                except Exception as e:
                    logging.error(e)
                    traceback.print_exc()
                    time.sleep(5)


class XlsxIbyFileD(XlsIbyFileD):
    def __init__(self, path, extension='xlsx'):
        super().__init__(path=path, extension=extension)

    def get_count(self, index, *args, **kwargs):
        all_line = 0
        workbook = openpyxl.load_workbook(self.gen_path_by_index(index))
        for idx, worksheet in enumerate(workbook):
            logging.info(f'sheet:{idx}:{worksheet.title}')
            all_line += worksheet.max_row
        return all_line

    def get_data(self, index):
        workbook = openpyxl.load_workbook(self.gen_path_by_index(index))  # 文件路径
        # 获取所有sheet的名字
        for idx, worksheet in enumerate(workbook):
            logging.info(f'sheet:{idx}:{worksheet.title}')
            nrows = worksheet.max_row
            if not nrows:
                continue
            keys = {i: key.value for i, key in enumerate(worksheet[1])}
            for row in worksheet.iter_rows(min_row=2):
                data = {keys[i]: cell.value for i, cell in enumerate(row)}
                yield data

    def get_cols_name_set(self, index):
        workbook = openpyxl.load_workbook(self.gen_path_by_index(index))  # 文件路径
        # 获取所有sheet的名字
        cols_name = set()
        for idx, worksheet in enumerate(workbook):
            logging.info(f'sheet:{idx}:{worksheet.title}')
            nrows = worksheet.max_row
            if not nrows:
                continue
            keys = {i: key.value for i, key in enumerate(worksheet[1])}
            print(f'sheet:{idx}:{worksheet.title}, keys:{keys}')
            cols_name = cols_name | set(keys.values())
        cols_name.remove(None)
        cols_name_l = list(cols_name)
        cols_name_l.sort()
        return cols_name_l


class MongoDBD(object):
    def __init__(self, hosts="mongodb://localhost:27017/", database='test', batch_size=1000):
        self.hosts = hosts
        self.batch_size = batch_size
        self.database = database
        self.client = pymongo.MongoClient(self.hosts)
        # self.db_list = self.client.list_database_names()
        self.db = self.client[self.database]
        # self.db.authenticate(user, password)
        self.collection_list = self.db.collection_names()
        self.gridfs = gridfs.GridFS(self.db)

    def __repr__(self):
        return f'MongoDB:{self.hosts}/{self.database}'

    def get_data(self, index, *args, **kwargs):
        for d in self.db[index].find().batch_size(self.batch_size):
            d.pop('_id')
            yield d
            # print(list(type(v) for k, v in d.items()))
        #     ObjectId
        # return

    def save_data(self, index, data, *args, **kwargs):
        res = self.db[index].insert_many(data)
        return res.inserted_ids

    def get_indexes(self):
        return list(self.db.collection_names())

    def get_count(self, index, *args, **kwargs):
        return self.db[index].find().count()

    def create_index(self, index, data, pks='id'):
        # sql = """SET NAMES utf8mb4;"""
        try:
            self.db[index].rename(f'{index}_{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}_bak')
        except pymongo.errors.OperationFailure as e:
            if e.__str__() != 'source namespace does not exist':
                raise

    def gridfs_save_file(self, content):
        with open(content, 'rb') as f:
            data = f.read()
            return self.gridfs.put(data)

    def gridfs_get_file(self, id):
        gf = self.gridfs.get(ObjectId(id))
        return gf.read()

    def gridfs_get_filename(self):
        return self.gridfs.list()


class BaseClient(ABC):
    def __init__(self):
        self.host = None
        self.port = None
        self.database = None
        self.username = None
        self._conn = None
        self.cur_lock = Lock()

    @abc.abstractmethod
    def connect(self):
        raise NotImplementedError('TODO connect function')

    def reconnect(self):
        self._conn = self.connect()

    @property
    def conn(self):
        if not self._conn:
            self.reconnect()
        return self._conn

    @property
    def cur(self):
        return self.conn.cursor()

    @classmethod
    def gen_insert_sql(cls, table_name, data, duplicate_update=False, partition='',):
        keys = []
        values = []
        for k, v in data.items():
            keys.append(k)
            if isinstance(v, (tuple, set)):
                v = list(v)
            if isinstance(v, (dict, list)):
                v = json.dumps(v)
            values.append(v)
        sql = f'insert into {table_name} {partition} ({",".join("`{}`".format(k) for k in keys)}) ' \
              f'values({",".join(["%s"] * len(values))})' + (
                  f'ON DUPLICATE KEY UPDATE {",".join("%s=(%s)" % (k, k) for k in keys)}'
                  if duplicate_update else '')
        return sql, values

    def insert_data(self, table_name, data, duplicate_update=False, partition=''):
        sql, values = self.gen_insert_sql(table_name, data, duplicate_update, partition)
        return self.execute(sql, parms=values)

    def insert_many_by_values_list(self, sql, insert_rows: [], batch_size=4000):
        insert_num = int((len(insert_rows) + batch_size - 1) / batch_size)
        for i in range(insert_num):
            temp = insert_rows[i * batch_size:(i + 1) * batch_size:]
            logging.info('准备插入{}条数据'.format(len(temp)))
            try:
                self.executemany(sql, temp)
            except Exception as ex:
                logging.error(format_error(ex))

    def insert_many_by_data(self, table_name, insert_rows: [], batch_size=4000, duplicate_update=False, partition=''):
        if not insert_rows:
            return
        sql, values = self.gen_insert_sql(table_name, insert_rows[0], duplicate_update, partition)
        insert_num = int((len(insert_rows) + batch_size - 1) / batch_size)
        for i in range(insert_num):
            temp = insert_rows[i * batch_size:(i + 1) * batch_size:]
            logging.info('准备插入{}条数据'.format(len(temp)))
            try:
                self.executemany(sql, [tuple(v.values()) for v in temp])
            except Exception as ex:
                logging.error(format_error(ex))
                raise ex

    @classmethod
    def gen_partition_sql(cls, **kwargs):
        return "partition({})".format(', '.join(f'{k}={v.__repr__()}' for k, v in kwargs.items()))

    def insert_random_data(self, table_name):
        keys, types = self.get_tables_struct(table_name)
        values = [gen_pass(v) for v in types]
        data = {k: v for k, v in zip(keys, values)}
        self.insert_data(table_name, data)
        # self.insert_data_from_table_orc(table_name, data, tables_struct={k: t for k, t in zip(keys, types)})

    def insert_data_from_table_orc(self, table_name, data, tables_struct=None):
        tmp_table_name = self.create_tmp_table(table_name, tables_struct)
        self.insert_data(tmp_table_name, data)
        self.execute(f'INSERT INTO TABLE {table_name} SELECT * FROM {tmp_table_name}')
        self.drop_tmp_table(tmp_table_name)

    def create_table_group_users(self):
        data = {
            "id": 'bigint',
            "chat_id": 'bigint',
            # "group_id": 'bigint',
            "user_id": 'bigint',
            "timestamp": 'bigint',
            "identity": 'INT',
            "last_active_time": 'bigint',
            "in_group": 'INT'
        }
        self.create_table('group_users', data, partitioned={
            "group_id": "string",
            # "year": "int",
            # "month": "int",
        })
        # self.create_table('group_users_no_orc', data, stored='')

    def create_table(
            self, table_name, data,
            partitioned=None,
            # stored='''STORED AS ORC'''
            stored='''STORED AS parquet'''
            # COLLECTION ITEMS TERMINATED BY '\t'
            # MAP KEYS TERMINATED BY '='
    ):
        create_sql = f"""
        CREATE TABLE IF NOT EXISTS `{table_name}`  (
            {', '.join('`%s` %s' % (k, v) for k, v in data.items())}
        )
        {"partitioned by ({})".format(','.join(
            '%s %s' % (k, v) for k, v in partitioned.items())) if partitioned else ''}
        row format delimited fields terminated by '\001'  --csv分隔
        collection items terminated by '\002'  --数组分隔
        map keys terminated by '\003'  --map分隔
        lines terminated by '\n'  --换行分隔
        {'tblproperties("skip.header.line.count"="0")    --跳过文件行首0行' if not stored else ''}
        {stored}
        """
        print(create_sql)
        self.execute(create_sql)

    def drop_tmp_table(self, table_name):
        if table_name.find('no_orc') > 0:
            print(f'DROP TABLE {table_name}')
            self.execute(f'DROP TABLE {table_name}')
            print(f'ok  DROP TABLE {table_name}')

    def create_tmp_table(self, table_name, tables_struct=None):
        if not tables_struct:
            tables_struct = {k: v for k, v in zip(*self.get_tables_struct(table_name))}
        tmp_table_name = f'{table_name}_no_orc_tmp_{time.time()}'.replace(".", "_")
        self.create_table(tmp_table_name, tables_struct, stored='')
        return tmp_table_name

    def load_data(self, path, table_name, tmp_table=False, local=False, tables_struct=None, partition=''):
        # 将临时表中的数据导入到ORC表中, 建立临时表1s load临时表1s 删除临时表1s
        if tmp_table:
            table_name_tmp_table = self.create_tmp_table(table_name, tables_struct)
            try:
                # "load data inpath '/user/admin/data/titanic/train.csv' OVERWRITE into table tmp"
                self.load_data(
                    path, table_name_tmp_table,
                    tmp_table=False, local=local, tables_struct=tables_struct, partition=partition
                )
                self.execute(f"insert into table {table_name} {partition} select * from {table_name_tmp_table}")
            except Exception as e:
                logging.error(format_error(e))
            finally:
                self.drop_tmp_table(table_name_tmp_table)
        else:
            self.execute(f"load data{' local' if local else ''} inpath '{path}' into table {table_name} {partition}")

    @abc.abstractmethod
    def show_table_struct(self, table_name):
        """
        查看表结构
        """
        raise NotImplementedError('TODO show_table_struct function')

    # @abc.abstractmethod
    # def show_table_info(self, table_name):
    #     """
    #     查看表详细属性
    #     """
    #     raise NotImplementedError('TODO show_table_info function')

    @abc.abstractmethod
    def get_tables_struct(self, table_name):
        """
        获取表详细属性list
        """
        raise NotImplementedError('TODO get_tables_struct function')

    def show_tables(self):
        return self.execute('SHOW tables')

    @with_cur_lock()
    def executemany(self, cur, sql, parms=None, *args, **kwargs):
        cur.executemany(sql, parms, *args, **kwargs)
        return cur.fetchall()

    @with_cur_lock()
    def execute(self, cur, sql, parms=None, *args, **kwargs):
        print(sql)
        cur.execute(sql, parms, *args, **kwargs)
        return cur.fetchall()

    @with_cur_lock()
    def execute_iter(self, cur, sql, parms=None, *args, **kwargs):
        print(sql)
        cur.execute(sql, parms, *args, **kwargs)
        result = cur.fetchmany(1000)
        while result:
            for r in result:
                yield r
            result = cur.fetchmany(1000)

    @with_cur_lock()
    def execute_iter_d(self, cur, sql, parms=None, *args, **kwargs):
        print(sql)
        cur.set_stream_results(True, 1000)
        cur.execute(sql, parms, *args, **kwargs)
        result = cur.fetchmany(1000)
        while result:
            for r in result:
                yield dict(zip(cur._columns, r))
            result = cur.fetchmany(1000)


class ClickHouseD(BaseClient):
    min_datetime = datetime.datetime(1971, 1, 1, 0, 0, 0)
    max_datetime = datetime.datetime(2106, 1, 1, 0, 0, 0)
    default_types = {'String': '', 'Int': 0, 'Float': 0.0, 'DateTime': min_datetime}

    def __init__(self, host, port, user, password, database):
        super().__init__()
        self.host = host
        self.port = port
        self.user = user
        self._cur = None

        self.password = password
        self.database = database

    @property
    def cur(self):
        if not self._cur:
            self._cur = self.conn.cursor()
        return self._cur

    def __repr__(self):
        return f'ClickHouse:{self.host}'

    def connect(self):
        return clickhouse_connect(
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.database,
        )

    def get_tables_struct(self, table_name):
        vs_info = self.show_table_struct(table_name)
        keys = [v[0] for v in vs_info]
        types = [v[1] for v in vs_info]
        return keys, types

    def show_table_struct(self, table_name):
        """
        查看表结构
        """
        return self.execute(f'desc {table_name}')

    def cols_type(self, table_name):
        return {kv[0]: kv[1] for kv in self.show_table_struct(table_name)}

    def get_count(self, index, *args, **kwargs):
        if index.lower().strip().startswith('select '):
            index = f'({index.strip()})'
        return self.execute(f'select count(0) as c from  {index} as taSDFEWVempTABlesdfecH')[0][0]

    @classmethod
    def gen_insert_sql_csv(cls, table_name, data):
        keys = []
        values = []
        for k, v in data.items():
            keys.append(k)
            if isinstance(v, (tuple, set)):
                v = list(v)
            if isinstance(v, (dict, list)):
                v = json.dumps(v)
            values.append(v)
        sql = f'insert into {table_name} select {",".join(v.__repr__() for v in values)}'
        return sql, values

    @classmethod
    def gen_insert_sql_no_v(cls, table_name, data):
        keys = []
        values = []
        for k, v in data.items():
            keys.append(k)
            if isinstance(v, (tuple, set)):
                v = list(v)
            if isinstance(v, (dict, list)):
                v = json.dumps(v)
            values.append(v)
        sql = f'INSERT INTO {table_name} ({",".join("`{}`".format(k) for k in keys)}) VALUES'
        return sql, values

    @classmethod
    def format_data_range(cls, data):
        for k, v in data.items():
            # 不能用isinstance，isinstance会考虑datetime的父类date
            if type(data[k]) == datetime.date:
                if data[k] < cls.min_datetime.date() or data[k] > cls.max_datetime.date():
                    data[k] = cls.min_datetime.date()
                if getattr(data[k], 'read', None):
                    data[k] = data[k].read()
            elif type(data[k]) == datetime.datetime:
                if data[k] < cls.min_datetime or data[k] > cls.max_datetime:
                    data[k] = cls.min_datetime
                if getattr(data[k], 'read', None):
                    data[k] = data[k].read()
        return data

    def data_types_check(self, data, cols_types):
        """table_from和table_to数据类型不一致时使用，会降低性能, 默认False"""
        type_mapping = {'String': str, 'Float': float, 'Int': int}
        for k, v in data.items():
            if type(v).__name__ not in cols_types[k].lower():
                for type_k in type_mapping:
                    if cols_types[k].find(type_k) != -1:
                        try:
                            data[k] = type_mapping[type_k](v)
                        except Exception as e:
                            print(type(e), e)
                            data[k] = self.default_types[type_k]
        return data

    def save_data(self, index, data, windows=10000, *args, **kwargs):
        if not data:
            return
        insert_rows = [_ for _ in data if _ and any(_.values())]
        if not insert_rows:
            return
        sql, values = self.gen_insert_sql_no_v(index, insert_rows[0])
        insert_num = int((len(insert_rows) + windows - 1) / windows)
        for i in range(insert_num):
            temp = insert_rows[i * windows:(i + 1) * windows:]
            logging.info('准备插入{}条数据'.format(len(temp)))
            try:
                self.executemany(
                    sql,
                    [self.format_data_range(v) for v in temp])
            except Exception as ex:
                logging.error(format_error(ex))
                raise ex

    def insert_from_mysql(self, table_to, host, port, database, table, user, password):
        # 直接插入到clickhouse现有表
        sql = f"""insert into {table_to} SELECT * FROM mysql('{host}:{port}', '{database}', '{table}', '{user}', '{password}')"""
        self.execute(sql, types_check=True)

    def is_table_exist(self, index):
        try:
            if self.execute(f'select count(*) from {index}'):
                return True
        except Exception:
            return False

    def create_index(self, index, data, pks='id'):
        if self.is_table_exist(index):
            return
        sql = """create table if not exists {}(""".format(index)
        for name, value in data.items():
            if isinstance(value, int):
                _type = 'Int64'
            elif isinstance(value, str) or value is None:
                _type = 'String'
            elif isinstance(value, datetime.date):
                _type = 'DateTime'
            elif isinstance(value, float):
                _type = 'Float'
            else:
                print(value, type(value))
                raise

            sql += ('`' + name + '` ')
            sql += (_type + ',')

        sql = sql.strip(',') + ')'

        sql += """ENGINE = Memory();"""
        self.execute(sql)

    def set_default_data(self, index):
        ch_struct = self.get_tables_struct(index)
        default_value = []
        for col_type in ch_struct[1]:
            d_len = len(default_value)
            for t in self.default_types:
                if t in col_type:
                    default_value.append(self.default_types[t])
                    break
            if len(default_value) != d_len + 1:
                print(col_type)
                raise
        default_data = dict(zip(ch_struct[0], default_value))
        return default_data

    def show_create_table(self, index, distribution=False):
        if distribution:
            pass
        else:
            return self.execute(f'show create table {index}')[0][0].replace('\n', ' ')

    def get_data(self, index, *args, **kwargs):
        sub_sql = f"{index} {kwargs.pop('condition')}" if 'condition' in kwargs else index
        if index.lower().strip().startswith('select '):
            return self.execute_iter_d(sql=f'{sub_sql}', *args, **kwargs)
        else:
            return self.execute_iter_d(sql=f'select * from {sub_sql}', *args, **kwargs)

    def get_indexes(self):
        return [i[0] for i in self.execute(f'show tables from {self.database}')]

    def get_cols_type(self, index):
        return dict({row[0]: row[1] for row in self.show_table_struct(index)})


class ListD:
    def __init__(self, index='default', data=None):
        self.data = {index: data or []}

    def __repr__(self):
        return f'ListD:{list(self.data.keys())}'

    def create_index(self, index='default', *args, **kwargs):
        self.data[index] = []

    def get_data(self, index='default', *args, **kwargs):
        return self.data[index]

    def get_dict_kv(self, index='default', pkey='', *args, **kwargs):
        return {item.get(pkey, ''): item for item in self.data[index]}

    def get_dict_kv_set(self, index='default', pkey='', *args, **kwargs):
        res = defaultdict(set)
        for item in self.data[index]:
            res[item.get(pkey, '')].add(item)
        return res

    def save_data(self, index='default', data=[], *args, **kwargs):
        self.data[index].extend(data)

    def get_indexes(self):
        return list(self.data.keys())

    def get_count(self, index='default', *args, **kwargs):
        return len(self.data[index])


class OracleD(BaseClient):
    def __init__(self, user, password, dsn, database):
        super().__init__()
        self.user = user
        self.password = password
        self.dsn = dsn
        self.database = database

    def __repr__(self):
        return f'Oracle:{self.dsn}'

    def connect(self):
        from cx_Oracle import connect as oracle_connect
        return oracle_connect(
            user=self.user,
            password=self.password,
            dsn=self.dsn,
            encoding='UTF-8'
        )

    @with_cur_lock()
    def execute(self, cur, sql, batch=100000, *args, **kwargs):
        print(sql)
        cur.execute(sql, *args, **kwargs)
        return cur.fetchall()

    def get_column_name(self, index):
        sql = f"select column_name from user_tab_columns  where Table_Name='{index.upper()}'"
        result = ()
        for c in self.execute(sql):
            result = result + c
        return result

    def get_column_name1(self, index):
        sql = f"select column_name from all_tab_cols where table_name='{index.upper()}' and owner='{self.databases}'"
        return [c[0] for c in self.execute(sql)]

    def get_count(self, index, *args, **kwargs):
        sql = f'select count(*) from {self.database}.{index.upper()}'
        if 'condition' in kwargs:
            sql += kwargs['condition']
        return self.execute(sql)[0][0]

    @with_cur_lock()
    def _execute(self, cur, sql, values=None, batch=10000, **kwargs):
        from cx_Oracle import LOB as oracle_LOB, DatabaseError as oracle_DatabaseError
        print(sql)
        try:
            if values:
                cur.execute(sql, values or [])
            else:
                cur.execute(sql)

            def _fetch(_dbcur):
                try:
                    result = _dbcur.fetchmany(batch)
                except oracle_DatabaseError:
                    time.sleep(600)
                    result = run_task_auto_retry(_dbcur.fetchmany, kwargs={'batch': batch})
                while result:
                    for r in result:
                        yield r
                    try:
                        result = _dbcur.fetchmany(batch)
                    except oracle_DatabaseError:
                        time.sleep(600)
                        result = run_task_auto_retry(_dbcur.fetchmany, kwargs={'batch': batch})
            return _fetch(cur)
        except Exception as e:
            raise e

    def get_data(self, index, *args, **kwargs):
        cols = self.get_column_name1(index)
        sql = f"select {','.join(cols)} from {self.database}.{index}"
        if 'condition' in kwargs:
            sql += kwargs['condition']
        result = self._execute(sql)
        for r in result:
            yield dict(zip(tuple(cols), r))

    def get_tables_struct(self, table_name):
        vs_info = self.show_table_struct(table_name)
        keys = [v[0] for v in vs_info]
        types = [v[1] for v in vs_info]
        return keys, types

    def show_table_struct(self, table_name):
        """
        查看表结构
        """
        sql = f"select column_name,data_type from all_tab_cols where table_name='{table_name}' and owner='{self.databases}'"
        return self.execute(sql)

    def get_cols_type(self, index):
        return dict({row[0]: row[1] for row in self.show_table_struct(index)})



import abc


class BaseStore(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def count_data(self, data_type=None, *args, **kwargs):
        return NotImplementedError

    @abc.abstractmethod
    def list_data(self, data_type=None, *args, **kwargs):
        return NotImplementedError

    @abc.abstractmethod
    def check_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    @abc.abstractmethod
    def get_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    @abc.abstractmethod
    def save_data(self, position, data, data_type=None, *args, **kwargs):
        return NotImplementedError

    @abc.abstractmethod
    def delete_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    @abc.abstractmethod
    def get_position(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    @abc.abstractmethod
    def get_data_size(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    @abc.abstractmethod
    def check_self(self, *args, **kwargs):
        return NotImplementedError

    @abc.abstractmethod
    def save_self(self, *args, **kwargs):
        return NotImplementedError

    @abc.abstractmethod
    def free_self(self, *args, **kwargs):
        return NotImplementedError



from .utils import (
    ElasticSearchD, MySqlD, CsvD, JsonListD, ListD, ZipD, RarD, TxtD,
    XlsIbyFileD, XlsxIbyFileD, MongoDBD, SqlFileD, ClickHouseD, OracleD,
    secure_filename, get_realpath, get_line_num_fast, LogFormatter, ziputils, rarutils, f2r, rsa
)
from .task import (
    Migration, Migration2DB, open_log
)
